from libmuscle.communicator import Communicator, Endpoint, Message
from libmuscle.mcp.direct_client import DirectClient
from libmuscle.mcp.direct_server import DirectServer
from libmuscle.mcp.message import ClosePort, Message as MCPMessage
from libmuscle.port import Port

from ymmsl import Conduit, Identifier, Operator, Reference, Settings

import msgpack
import pytest
from unittest.mock import patch, MagicMock


def test_endpoint() -> None:
    kernel = Reference('test.kernel')
    index = [42]
    port = Identifier('out')
    slot = [2]

    endpoint = Endpoint(kernel, index, port, slot)
    assert endpoint.kernel == kernel
    assert endpoint.index == index
    assert endpoint.port == port
    assert endpoint.slot == slot

    assert str(endpoint) == 'test.kernel[42].out[2]'


def test_endpoint_instance() -> None:
    endpoint = Endpoint(Reference('test.kernel'), [42], Identifier('port'),
                        [2])
    assert endpoint.instance() == 'test.kernel[42]'

    endpoint2 = Endpoint(Reference('test.kernel'), [], Identifier('port'), [])
    assert endpoint2.instance() == 'test.kernel'

    endpoint3 = Endpoint(Reference('test.kernel'), [], Identifier('port'), [3])
    assert endpoint3.instance() == 'test.kernel'


@pytest.fixture
def communicator() -> Communicator:
    # Creating a Communicator will start servers, and some servers are not
    # fork-compatible. This then crashes the later integration tests, which
    # fork. So we disable everything except DirectServer.
    with patch('libmuscle.communicator.server_types', [DirectServer]):
        instance_id = Reference('kernel')
        communicator = Communicator(instance_id, [13], None, MagicMock())
        communicator._peer_manager = MagicMock()
        pm = communicator._peer_manager
        pm.is_connected.return_value = True

        def gpp(x) -> Reference:
            if 'out' in str(x):
                return Reference('in')
            return Reference('out')

        pm.get_peer_port = gpp

        pm.get_peer_dims.return_value = []
        pm.get_peer_locations.return_value = ['direct:test']

        def gpe(p, s) -> Reference:
            endpoint = MagicMock()
            endpoint.instance.return_value = Reference('other')
            if 'out' in str(p):
                endpoint.ref.return_value = Reference('other.in[13]')
            else:
                endpoint.ref.return_value = Reference('other.out')
            return endpoint

        pm.get_peer_endpoint = gpe

        communicator._ports = {
                'out': Port('out', Operator.O_I, False, True, 1, []),
                'in': Port('in', Operator.S, False, True, 1, [])}
        yield communicator
        communicator.shutdown()


@pytest.fixture
def communicator2() -> Communicator:
    with patch('libmuscle.communicator.server_types', [DirectServer]):
        instance_id = Reference('other')
        communicator = Communicator(instance_id, [], None, MagicMock())
        communicator._peer_manager = MagicMock()
        pm = communicator._peer_manager
        pm.is_connected.return_value = True

        def gpp(x: Reference) -> Reference:
            if 'out' in str(x):
                return Reference('in')
            return Reference('out')

        pm.get_peer_port = gpp

        pm.get_peer_dims.return_value = []
        pm.get_peer_locations.return_value = ['direct:test']

        def gpe(p, s) -> Reference:
            endpoint = MagicMock()
            endpoint.instance.return_value = Reference('kernel[13]')
            if 'out' in str(p):
                endpoint.ref.return_value = Reference('kernel[13].in')
            else:
                endpoint.ref.return_value = Reference('kernel[13].out')
            return endpoint

        pm.get_peer_endpoint = gpe

        communicator._ports = {
                'out': Port('out', Operator.O_I, True, True, 0, [20]),
                'in': Port('in', Operator.S, True, True, 0, [20])}
        yield communicator
        communicator.shutdown()


@pytest.fixture
def communicator3() -> Communicator:
    with patch('libmuscle.communicator.server_types', [DirectServer]):
        instance_id = Reference('kernel')
        communicator = Communicator(instance_id, [], None, MagicMock())
        communicator._peer_manager = MagicMock()
        pm = communicator._peer_manager
        pm.is_connected.return_value = True

        def gpp(x: Reference) -> Reference:
            if 'out' in str(x):
                return Reference('in')
            return Reference('out')

        pm.get_peer_port = gpp

        pm.get_peer_dims.return_value = []
        pm.get_peer_locations.return_value = ['direct:test']

        def gpe(p, s) -> Reference:
            endpoint = MagicMock()
            endpoint.instance.return_value = Reference('other')
            if 'out' in str(p):
                endpoint.ref.return_value = Reference('other.in[13]')
            else:
                endpoint.ref.return_value = Reference('other.out[13]')
            return endpoint

        pm.get_peer_endpoint = gpe

        communicator._ports = {
                'out': Port('out', Operator.O_I, True, True, 0, []),
                'in': Port('in', Operator.S, True, True, 0, [])}
        yield communicator
        communicator.shutdown()


def test_create_communicator(communicator) -> None:
    assert str(communicator._kernel) == 'kernel'
    assert communicator._index == [13]
    assert len(communicator._servers) == 1
    assert communicator._clients == {}
    assert communicator._post_office._outboxes == {}


def test_get_locations(communicator) -> None:
    assert len(communicator.get_locations()) == 1
    assert communicator.get_locations()[0].startswith('direct:')


def test_connect() -> None:
    ref = Reference

    instance_id = Reference('kernel')
    conduits = [Conduit('kernel.out', 'other.in'),
                Conduit('other.out', 'kernel.in')]
    peer_dims = {ref('other'): [1]}
    peer_locations = {ref('other'): ['direct:test']}

    with patch('libmuscle.communicator.PeerManager') as pm_init:
        communicator = Communicator(instance_id, [13], None, MagicMock())

        communicator.connect(conduits, peer_dims, peer_locations)

        pm_init.assert_called_with(instance_id, [13], conduits, peer_dims,
                                   peer_locations)

        # check inferred ports
        ports = communicator._ports
        communicator.shutdown()
        assert ports['in'].name == Identifier('in')
        assert ports['in'].operator == Operator.F_INIT
        assert ports['in']._length is None

        assert ports['out'].name == Identifier('out')
        assert ports['out'].operator == Operator.O_F
        assert ports['out']._length is None


def test_connect_vector_ports(communicator) -> None:
    ref = Reference

    communicator._declared_ports = {
            Operator.F_INIT: ['in[]'],
            Operator.O_F: ['out1', 'out2[]']}

    conduits = [Conduit('other1.out', 'kernel.in'),
                Conduit('kernel.out1', 'other.in'),
                Conduit('kernel.out2', 'other3.in')]
    peer_dims = {
            ref('other1'): [20, 7],
            ref('other'): [25],
            ref('other3'): [20]}
    peer_locations = {
            ref('other'): ['direct:test'],
            ref('other1'): ['direct:test1'],
            ref('other3'): ['direct:test3']}

    communicator.connect(conduits, peer_dims, peer_locations)

    ports = communicator._ports
    assert ports['in'].name == Identifier('in')
    assert ports['in'].operator == Operator.F_INIT
    assert ports['in']._length == 7
    assert ports['in']._is_resizable is False

    assert ports['out1'].name == Identifier('out1')
    assert ports['out1'].operator == Operator.O_F
    assert ports['out1']._length is None

    assert ports['out2'].name == Identifier('out2')
    assert ports['out2'].operator == Operator.O_F
    assert ports['out2']._length == 0
    assert ports['out2']._is_resizable is True


def test_connect_multidimensional_ports(communicator) -> None:
    ref = Reference

    communicator._declared_ports = {
            Operator.F_INIT: ['in[][]']}

    conduits = [Conduit(ref('other.out'), ref('kernel.in'))]
    peer_dims = {ref('other'): [20, 7, 30]}
    peer_locations = {ref('other'): ['direct:test']}
    with pytest.raises(ValueError):
        communicator.connect(conduits, peer_dims, peer_locations)


def test_connect_inferred_ports(communicator) -> None:
    ref = Reference

    communicator._declared_ports = None

    conduits = [Conduit('other1.out', 'kernel.in'),
                Conduit('kernel.out1', 'other.in'),
                Conduit('kernel.out3', 'other2.in')]
    peer_dims = {
            ref('other1'): [20, 7],
            ref('other'): [25],
            ref('other2'): []}
    peer_locations = {
            ref('other'): ['direct:test'],
            ref('other1'): ['direct:test1'],
            ref('other2'): ['direct:test2']}

    communicator.connect(conduits, peer_dims, peer_locations)

    ports = communicator._ports
    assert ports['in'].name == Identifier('in')
    assert ports['in'].operator == Operator.F_INIT
    assert ports['in']._length == 7
    assert ports['in']._is_resizable is False

    assert ports['out1'].name == Identifier('out1')
    assert ports['out1'].operator == Operator.O_F
    assert ports['out1']._length is None

    assert ports['out3'].name == Identifier('out3')
    assert ports['out3'].operator == Operator.O_F
    assert ports['out3']._length is None


def test_send_message(communicator, message) -> None:
    communicator.send_message('out', message)

    assert 'other.in[13]' in communicator._post_office._outboxes
    msg_bytes = communicator._post_office._outboxes[
            'other.in[13]']._Outbox__queue.get()
    msg = MCPMessage.from_bytes(msg_bytes)
    assert msg.sender == 'kernel[13].out'
    assert msg.receiver == 'other.in[13]'
    assert msg.timestamp == 0.0
    assert msg.next_timestamp is None
    assert msg.settings_overlay == Settings()
    assert msg.data == b'test'


def test_send_on_disconnected_port(communicator, message) -> None:
    communicator._peer_manager.is_connected.return_value = False
    communicator.send_message('not_connected', message)


def test_send_on_invalid_port(communicator, message) -> None:
    with pytest.raises(ValueError):
        communicator.send_message('[$Invalid_id', message)


def test_send_msgpack(communicator, message2) -> None:
    communicator.send_message('out', message2)

    assert 'other.in[13]' in communicator._post_office._outboxes
    msg_bytes = communicator._post_office._outboxes[
            'other.in[13]']._Outbox__queue.get()
    msg = MCPMessage.from_bytes(msg_bytes)
    assert msg.sender == 'kernel[13].out'
    assert msg.receiver == 'other.in[13]'
    assert msg.settings_overlay == Settings()
    assert msg.data == {'test': 17}


def test_send_message_with_slot(communicator2, message) -> None:
    communicator2.send_message('out', message, 13)

    assert 'kernel[13].in' in \
        communicator2._post_office._outboxes
    msg_bytes = communicator2._post_office._outboxes[
            'kernel[13].in']._Outbox__queue.get()
    msg = MCPMessage.from_bytes(msg_bytes)
    assert msg.sender == 'other.out[13]'
    assert msg.receiver == 'kernel[13].in'
    assert msg.settings_overlay == Settings()
    assert msg.data == b'test'


def test_send_message_resizable(communicator3, message) -> None:
    with pytest.raises(RuntimeError):
        communicator3.send_message('out', message, 13)

    communicator3.get_port('out').set_length(20)
    communicator3.send_message('out', message, 13)

    assert 'other.in[13]' in communicator3._post_office._outboxes
    msg_bytes = communicator3._post_office._outboxes[
            'other.in[13]']._Outbox__queue.get()
    msg = MCPMessage.from_bytes(msg_bytes)
    assert msg.sender == 'kernel.out[13]'
    assert msg.receiver == 'other.in[13]'
    assert msg.port_length == 20


def test_send_message_with_settings(communicator, message) -> None:
    message.settings['test2'] = 'testing'
    communicator.send_message('out', message)

    assert 'other.in[13]' in communicator._post_office._outboxes
    msg_bytes = communicator._post_office._outboxes[
            'other.in[13]']._Outbox__queue.get()
    msg = MCPMessage.from_bytes(msg_bytes)
    assert msg.sender == 'kernel[13].out'
    assert msg.receiver == 'other.in[13]'
    assert msg.settings_overlay.as_ordered_dict() == {'test2': 'testing'}
    assert msg.data == b'test'


def test_send_settings(communicator, message) -> None:
    message.data = Settings()
    message.data['test1'] = 'testing'
    communicator.send_message('out', message)

    assert 'other.in[13]' in communicator._post_office._outboxes
    msg_bytes = communicator._post_office._outboxes[
            'other.in[13]']._Outbox__queue.get()
    msg = MCPMessage.from_bytes(msg_bytes)
    assert msg.sender == 'kernel[13].out'
    assert msg.receiver == 'other.in[13]'
    assert msg.settings_overlay == Settings()
    assert msg.data == Settings({'test1': 'testing'})


def test_close_port(communicator) -> None:
    communicator.close_port('out')

    assert 'other.in[13]' in communicator._post_office._outboxes
    msg_bytes = communicator._post_office._outboxes[
            'other.in[13]']._Outbox__queue.get()
    msg = MCPMessage.from_bytes(msg_bytes)
    assert msg.sender == 'kernel[13].out'
    assert msg.receiver == 'other.in[13]'
    assert msg.timestamp == float('inf')
    assert msg.next_timestamp is None
    assert msg.settings_overlay == Settings()
    assert isinstance(msg.data, ClosePort)


def test_receive_message(communicator) -> None:
    client_mock = MagicMock()
    client_mock.receive.return_value = MCPMessage(
            Reference('other.out[13]'), Reference('kernel[13].in'),
            None, 0.0, None, Settings({'test1': 12}),
            b'test').encoded()
    get_client_mock = MagicMock(return_value=client_mock)
    communicator._Communicator__get_client = get_client_mock
    communicator._profiler = MagicMock()

    msg = communicator.receive_message('in')

    get_client_mock.assert_called_with(Reference('other'))
    client_mock.receive.assert_called_with(Reference('kernel[13].in'))
    assert msg.data == b'test'
    assert msg.settings['test1'] == 12


def test_receive_message_default(communicator) -> None:
    communicator._peer_manager.is_connected.return_value = False
    default_msg = Message(3.0, 4.0, 'test', Settings())
    msg = communicator.receive_message('not_connected', default=default_msg)
    assert msg.timestamp == 3.0
    assert msg.next_timestamp == 4.0
    assert msg.data == 'test'
    assert len(msg.settings) == 0


def test_receive_message_no_default(communicator) -> None:
    communicator._peer_manager.is_connected.return_value = False
    with pytest.raises(RuntimeError):
        communicator.receive_message('not_connected')


def test_receive_on_invalid_port(communicator) -> None:
    with pytest.raises(ValueError):
        communicator.receive_message('@$Invalid_id')


def test_receive_msgpack(communicator) -> None:
    client_mock = MagicMock()
    client_mock.receive.return_value = MCPMessage(
            Reference('other.out[13]'), Reference('kernel[13].in'),
            None, 0.0, None, Settings({'test1': 12}),
            {'test': 13}).encoded()
    get_client_mock = MagicMock(return_value=client_mock)
    communicator._Communicator__get_client = get_client_mock
    communicator._profiler = MagicMock()

    msg = communicator.receive_message('in')

    get_client_mock.assert_called_with(Reference('other'))
    client_mock.receive.assert_called_with(Reference('kernel[13].in'))
    assert msg.data == {'test': 13}


def test_receive_with_slot(communicator2) -> None:
    client_mock = MagicMock()
    client_mock.receive.return_value = MCPMessage(
            Reference('kernel[13].out'), Reference('other.in[13]'),
            None, 0.0, None, Settings({'test': 'testing'}),
            b'test').encoded()
    get_client_mock = MagicMock(return_value=client_mock)
    communicator2._Communicator__get_client = get_client_mock
    communicator2._profiler = MagicMock()

    msg = communicator2.receive_message('in', 13)

    get_client_mock.assert_called_with(Reference('kernel[13]'))
    client_mock.receive.assert_called_with(Reference('other.in[13]'))
    assert msg.data == b'test'
    assert msg.settings['test'] == 'testing'


def test_receive_message_resizable(communicator3) -> None:
    client_mock = MagicMock()
    client_mock.receive.return_value = MCPMessage(
            Reference('other.out[13]'), Reference('kernel.in[13]'),
            20, 0.0, None, Settings({'test': 'testing'}),
            b'test').encoded()
    get_client_mock = MagicMock(return_value=client_mock)
    communicator3._Communicator__get_client = get_client_mock
    communicator3._profiler = MagicMock()

    msg = communicator3.receive_message('in', 13)

    get_client_mock.assert_called_with(Reference('other'))
    client_mock.receive.assert_called_with(Reference('kernel.in[13]'))
    assert msg.data == b'test'
    assert communicator3.get_port('in').get_length() == 20


def test_receive_with_settings(communicator) -> None:
    client_mock = MagicMock()
    client_mock.receive.return_value = MCPMessage(
            Reference('other.out[13]'), Reference('kernel[13].in'),
            None, 0.0, None, Settings({'test2': 3.1}),
            b'test').encoded()
    get_client_mock = MagicMock(return_value=client_mock)
    communicator._Communicator__get_client = get_client_mock
    communicator._profiler = MagicMock()

    msg = communicator.receive_message('in')

    get_client_mock.assert_called_with(Reference('other'))
    client_mock.receive.assert_called_with(Reference('kernel[13].in'))
    assert msg.data == b'test'
    assert msg.settings['test2'] == 3.1


def test_receive_msgpack_with_slot_and_settings(communicator2) -> None:
    client_mock = MagicMock()
    client_mock.receive.return_value = MCPMessage(
            Reference('kernel[13].out'), Reference('other.in[13]'),
            None, 0.0, 1.0,
            Settings({'test': 'testing'}), 'test').encoded()
    get_client_mock = MagicMock(return_value=client_mock)
    communicator2._Communicator__get_client = get_client_mock
    communicator2._profiler = MagicMock()

    msg = communicator2.receive_message('in', 13)

    get_client_mock.assert_called_with(Reference('kernel[13]'))
    client_mock.receive.assert_called_with(Reference('other.in[13]'))
    assert msg.data == 'test'
    assert msg.settings['test'] == 'testing'


def test_receive_settings(communicator) -> None:
    client_mock = MagicMock()
    client_mock.receive.return_value = MCPMessage(
            Reference('other.out[13]'), Reference('kernel[13].in'),
            None, 0.0, None, Settings({'test1': 12}),
            Settings({'test': 13})).encoded()
    get_client_mock = MagicMock(return_value=client_mock)
    communicator._Communicator__get_client = get_client_mock
    communicator._profiler = MagicMock()

    msg = communicator.receive_message('in')

    get_client_mock.assert_called_with(Reference('other'))
    client_mock.receive.assert_called_with(Reference('kernel[13].in'))
    assert isinstance(msg.data, Settings)
    assert msg.data['test'] == 13


def test_receive_close_port(communicator) -> None:
    client_mock = MagicMock()
    client_mock.receive.return_value = MCPMessage(
            Reference('other.out[13]'), Reference('kernel[13].in'),
            None, 0.0, None, Settings(), ClosePort()).encoded()
    get_client_mock = MagicMock(return_value=client_mock)
    communicator._Communicator__get_client = get_client_mock
    communicator._profiler = MagicMock()

    msg = communicator.receive_message('in')

    assert isinstance(msg.data, ClosePort)


def test_get_message(communicator, message) -> None:
    communicator.send_message('out', message)
    ref_message = MCPMessage(
            Reference('kernel[13].out'), Reference('other.in[13]'),
            None, 0.0, None, Settings(), b'test').encoded()
    assert communicator._post_office.get_message(
            'other.in[13]') == ref_message


@patch('libmuscle.mcp.direct_client.registered_servers')
def test_get_client(mock_servers, communicator) -> None:
    mock_servers.__contains__.return_value = True
    client = communicator._Communicator__get_client(Reference('other'))
    mock_servers.__contains__.assert_called_with('test')
    assert isinstance(client, DirectClient)

    client2 = communicator._Communicator__get_client(Reference('other'))
    assert client == client2

    gpl = communicator._peer_manager.get_peer_locations
    gpl.return_value = ['non_existent:test']
    with pytest.raises(RuntimeError):
        communicator._Communicator__get_client(Reference('other2'))



import multiprocessing as mp
import time
from ymmsl import Reference

import pytest

from libmuscle.mcp.pipe_client import PipeClient
from libmuscle.mcp.pipe_server import PipeServer
import libmuscle.mcp.pipe_multiplexer as mux
from libmuscle.mcp.message import Message


def run_server(instance_id, receiver, post_office):
    server = PipeServer(instance_id, post_office)
    assert server.get_location().endswith('/{}'.format(instance_id))

    # wait for message to be sent
    while post_office.outboxes[receiver]._Outbox__queue.qsize() > 0:
        time.sleep(0.1)
    server.close()


def run_client(instance_id, server_location, receiver, message):
    assert PipeClient.can_connect_to(server_location)
    client = PipeClient(instance_id, server_location)

    message2 = client.receive(receiver)
    assert message.sender == message2.sender
    assert message.receiver == message2.receiver
    assert message.port_length == message2.port_length
    assert message.timestamp == message2.timestamp
    assert message.next_timestamp == message2.next_timestamp
    assert message.settings_overlay == message2.settings_overlay
    assert message.data == message2.data

    client.close()
    PipeClient.shutdown(instance_id)


def test_send_receive(receiver, post_office):
    message = Message(Reference('test_sender.test_port'), receiver,
                      None, 0.0, 1.0, bytes(), 'message'.encode('utf-8'))

    # prepare post office, it's about to get forked
    post_office.outboxes[receiver].deposit(message)

    # create server in separate process
    sender_instance_id = Reference('test_sender')
    mux.add_instance(sender_instance_id)
    server_proc = mp.Process(target=run_server,
                             args=(sender_instance_id, receiver, post_office),
                             name='PipeServer')
    server_proc.start()
    mux.close_instance_ends(sender_instance_id)

    # create client in separate process
    server_location = mux.get_address_for(sender_instance_id)

    recv_instance_id = Reference('test_receiver')
    mux.add_instance(recv_instance_id)
    client_proc = mp.Process(target=run_client,
                             args=(recv_instance_id, server_location, receiver,
                                   message),
                             name='PipeClient')
    client_proc.start()
    mux.close_instance_ends(recv_instance_id)

    # service connection requests
    mux.run()

    # shut down
    client_proc.join()
    server_proc.join()



from concurrent import futures
import logging
import time
import socket
from typing import cast, Generator, List

import grpc
from ymmsl import Reference, Settings

from libmuscle.port import port_from_grpc
from libmuscle.logging import LogLevel, Timestamp
from libmuscle.manager.instance_registry import InstanceRegistry
from libmuscle.manager.logger import Logger
from libmuscle.manager.topology_store import TopologyStore
from libmuscle.operator import operator_from_grpc
from libmuscle.util import (conduit_to_grpc, generate_indices,
                            instance_indices, instance_to_kernel)

import muscle_manager_protocol.muscle_manager_protocol_pb2 as mmp
import muscle_manager_protocol.muscle_manager_protocol_pb2_grpc as mmp_grpc


class MMPServicer(mmp_grpc.MuscleManagerServicer):
    """The MUSCLE Manager Protocol server.

    This class handles incoming requests from the instances comprising \
    the multiscale simulation to be executed.

    Args:
        logger: The Logger component to log messages to.
        settings: The global settings to serve to instances.
        instance_registry: The database for instances.
        topology_store: Keeps track of how to connect things.
    """
    def __init__(
            self,
            logger: Logger,
            settings: Settings,
            instance_registry: InstanceRegistry,
            topology_store: TopologyStore
            ) -> None:
        self.__logger = logger
        self.__settings = settings
        self.__instance_registry = instance_registry
        self.__topology_store = topology_store

    def SubmitLogMessage(
            self,
            request: mmp.LogMessage,
            context: grpc.ServicerContext
            ) -> mmp.LogResult:
        """Forwards a submitted log message to the Logger."""
        self.__logger.log_message(
                request.instance_id,
                Timestamp.from_grpc(request.timestamp),
                LogLevel.from_grpc(request.level),
                request.text)
        return mmp.LogResult()

    def SubmitProfileEvents(
            self,
            request: mmp.Profile,
            context: grpc.ServicerContext
            ) -> mmp.ProfileResult:
        """Forwards a submitted log message to the ProfilingStore."""
        # TODO: store
        return mmp.ProfileResult()

    def RequestSettings(
            self,
            request: mmp.SettingsRequest,
            context: grpc.ServicerContext
            ) -> mmp.SettingsResult:
        """Returns the central base settings."""
        settings = list()   # type: List[mmp.Setting]
        for name, value in self.__settings.items():
            if isinstance(value, str):
                setting = mmp.Setting(
                        name=str(name),
                        value_type=mmp.SETTING_VALUE_TYPE_STRING,
                        value_string=value)
            elif isinstance(value, bool):
                # a bool is an int in Python, so this needs to go before the
                # branch for int
                setting = mmp.Setting(
                        name=str(name),
                        value_type=mmp.SETTING_VALUE_TYPE_BOOL,
                        value_bool=value)
            elif isinstance(value, int):
                setting = mmp.Setting(
                        name=str(name),
                        value_type=mmp.SETTING_VALUE_TYPE_INT,
                        value_int=value)
            elif isinstance(value, float):
                setting = mmp.Setting(
                        name=str(name),
                        value_type=mmp.SETTING_VALUE_TYPE_FLOAT,
                        value_float=value)
            elif isinstance(value, list):
                if len(value) == 0 or isinstance(value[0], float):
                    value = cast(List[float], value)
                    mmp_values = mmp.ListOfDouble(values=value)
                    setting = mmp.Setting(
                            name=str(name),
                            value_type=mmp.SETTING_VALUE_TYPE_LIST_FLOAT,
                            value_list_float=mmp_values)
                elif isinstance(value[0], list):
                    value = cast(List[List[float]], value)
                    rows = list()
                    for row in value:
                        mmp_row = mmp.ListOfDouble(values=row)
                        rows.append(mmp_row)
                    mmp_rows = mmp.ListOfListOfDouble(values=rows)

                    LLF = mmp.SETTING_VALUE_TYPE_LIST_LIST_FLOAT
                    setting = mmp.Setting(
                            name=str(name),
                            value_type=LLF,
                            value_list_list_float=mmp_rows)
            settings.append(setting)

        return mmp.SettingsResult(setting_values=settings)

    def RegisterInstance(
            self,
            request: mmp.RegistrationRequest,
            context: grpc.ServicerContext
            ) -> mmp.RegistrationResult:
        """Handles an instance registration request."""
        try:
            ports = list(map(port_from_grpc, request.ports))
            self.__instance_registry.add(
                    Reference(str(request.instance_name)),
                    list(request.network_locations),
                    ports)
            self.__log(LogLevel.INFO, 'Registered instance {}'.format(
                        request.instance_name))
            return mmp.RegistrationResult(status=mmp.RESULT_STATUS_SUCCESS)
        except ValueError as e:
            return mmp.RegistrationResult(
                    status=mmp.RESULT_STATUS_ERROR,
                    error_message=('An instance with name {} was already'
                                   ' registered').format(
                                           request.instance_name))

    def RequestPeers(
            self,
            request: mmp.PeerRequest,
            context: grpc.ServicerContext
            ) -> mmp.PeerResult:
        """Handles a peer request."""
        # get info from yMMSL
        instance = Reference(request.instance_name)
        kernel = instance_to_kernel(instance)
        if not self.__topology_store.has_kernel(kernel):
            return mmp.PeerResult(status=mmp.RESULT_STATUS_ERROR,
                                  error_message='Unknown kernel {}'.format(
                                      kernel))

        conduits = self.__topology_store.get_conduits(kernel)
        mmp_conduits = [conduit_to_grpc(c) for c in conduits]

        peer_dims = self.__topology_store.get_peer_dimensions(kernel)
        mmp_dimensions = [
                mmp.PeerResult.PeerDimensions(peer_name=str(name),
                                              dimensions=dims)
                for name, dims in peer_dims.items()]

        # generate instances
        peer_instances = self.__generate_peer_instances(instance)
        try:
            instance_locations = [
                    mmp.PeerResult.PeerLocations(
                        instance_name=str(instance),
                        locations=self.__instance_registry.get_locations(
                            instance))
                    for instance in peer_instances]
        except KeyError as e:
            return mmp.PeerResult(status=mmp.RESULT_STATUS_PENDING,
                                  error_message='Waiting for kernel {}'.format(
                                      e.args[0]))

        self.__log(LogLevel.INFO, 'Sent peers to {}'.format(
                    request.instance_name))
        return mmp.PeerResult(
                status=mmp.RESULT_STATUS_SUCCESS,
                conduits=mmp_conduits,
                peer_dimensions=mmp_dimensions,
                peer_locations=instance_locations)

    def DeregisterInstance(self, request: mmp.DeregistrationRequest,
                           context: grpc.ServicerContext
                           ) -> mmp.DeregistrationResult:
        """Handles an instance deregistration request."""
        try:
            self.__instance_registry.remove(Reference(request.instance_name))
            self.__log(LogLevel.INFO, 'Deregistered instance {}'.format(
                    request.instance_name))
            return mmp.DeregistrationResult(status=mmp.RESULT_STATUS_SUCCESS)
        except ValueError as e:
            return mmp.DeregistrationResult(
                    status=mmp.RESULT_STATUS_ERROR,
                    error_message=('No instance with name {} was registered'
                                   ).format(request.instance_name))

    def __generate_peer_instances(self, instance: Reference
                                  ) -> Generator[Reference, None, None]:
        """Generates the names of all peer instances of an instance.

        Args:
            instance: The instance whose peers to generate.

        Yields:
            All peer instance identifiers.
        """
        kernel = instance_to_kernel(instance)
        indices = instance_indices(instance)
        dims = self.__topology_store.kernel_dimensions[kernel]
        all_peer_dims = self.__topology_store.get_peer_dimensions(kernel)
        for peer, peer_dims in all_peer_dims.items():
            base = peer
            for i in range(min(len(dims), len(peer_dims))):
                base += indices[i]

            if dims >= peer_dims:
                yield base
            else:

                for peer_indices in generate_indices(peer_dims[len(dims):]):
                    yield base + peer_indices

    def __log(self, level: LogLevel, msg: str) -> None:
        """Logs a message to the log file.

        Args:
            level: The level to log at.
            msg: The message to log.
        """
        self.__logger.log_message('muscle3_manager', Timestamp(time.time()),
                                  level, msg)


class MMPServer():
    """The MUSCLE Manager Protocol server.

    This class accepts connections from the instances comprising \
    the multiscale model to be executed, and services them using an \
    MMPServicer.
    """
    def __init__(
            self,
            logger: Logger,
            settings: Settings,
            instance_registry: InstanceRegistry,
            topology_store: TopologyStore
            ) -> None:
        self.__instance_registry = instance_registry
        self.__servicer = MMPServicer(logger, settings, instance_registry,
                                      topology_store)
        self.__server = grpc.server(futures.ThreadPoolExecutor())
        mmp_grpc.add_MuscleManagerServicer_to_server(  # type: ignore
                self.__servicer, self.__server)
        self.__server.add_insecure_port('[::]:9000')
        self.__server.start()

    def get_location(self) -> str:
        """Returns this server's network location.

        This is a string of the form <hostname>:<port>.
        """
        hostname = socket.getfqdn()
        return '{}:{}'.format(hostname, 9000)

    def wait(self) -> None:
        """Waits for the server to finish.

        The server will shut down after every instance has been
        registered and deregistered again.
        """
        self.__instance_registry.wait()
        time.sleep(1)
        self.__server.stop(5)

    def stop(self) -> None:
        """Stops the server.

        This makes the server stop serving requests, and shuts down its
        background threads.
        """
        self.__server.stop(0)



from google.protobuf.timestamp_pb2 import Timestamp
from ymmsl import Operator, Reference

from libmuscle.manager.mmp_server import MMPServicer
import muscle_manager_protocol.muscle_manager_protocol_pb2 as mmp


def test_create_servicer(logger, settings, instance_registry,
                         topology_store):
    MMPServicer(logger, settings, instance_registry, topology_store)


def test_log_message(mmp_servicer, caplog):
    timestamp = Timestamp()
    timestamp.FromJsonString("1970-01-01T00:00:00.000Z")
    message = mmp.LogMessage(
            instance_id='test_instance_id',
            timestamp=timestamp,
            level=mmp.LOG_LEVEL_WARNING,
            text='Testing log message')
    result = mmp_servicer.SubmitLogMessage(message, None)
    assert isinstance(result, mmp.LogResult)
    assert caplog.records[0].name == 'test_instance_id'
    assert caplog.records[0].time_stamp == '1970-01-01T00:00:00Z'
    assert caplog.records[0].levelname == 'WARNING'
    assert caplog.records[0].message == (
            'Testing log message')


def test_request_settings(settings, mmp_servicer):
    request = mmp.SettingsRequest()
    result = mmp_servicer.RequestSettings(request, None)

    assert len(result.setting_values) == 0

    settings['test1'] = 13
    settings['test2'] = 12.3
    settings['test3'] = 'testing'
    settings['test4'] = True
    settings['test5'] = [2.3, 7.4]
    settings['test6'] = [[1.0, 2.0], [2.0, 1.0]]

    result = mmp_servicer.RequestSettings(request, None)
    assert len(result.setting_values) == 6

    result_dict = dict()
    for mmp_setting in result.setting_values:
        result_dict[mmp_setting.name] = (
                mmp_setting.value_type, mmp_setting.value_string,
                mmp_setting.value_int, mmp_setting.value_float,
                mmp_setting.value_bool, mmp_setting.value_list_float,
                mmp_setting.value_list_list_float)

    no_list = mmp.ListOfDouble()
    no_list2 = mmp.ListOfListOfDouble()
    assert result_dict['test1'] == (
            mmp.SETTING_VALUE_TYPE_INT, '', 13, 0.0, False, no_list,
            no_list2)
    assert result_dict['test2'] == (
            mmp.SETTING_VALUE_TYPE_FLOAT, '', 0, 12.3, False, no_list,
            no_list2)
    assert result_dict['test3'] == (
            mmp.SETTING_VALUE_TYPE_STRING, 'testing', 0, 0.0, False, no_list,
            no_list2)
    assert result_dict['test4'] == (
            mmp.SETTING_VALUE_TYPE_BOOL, '', 0, 0.0, True, no_list, no_list2)
    assert result_dict['test5'] == (
            mmp.SETTING_VALUE_TYPE_LIST_FLOAT, '', 0, 0.0, False,
            mmp.ListOfDouble(values=[2.3, 7.4]), no_list2)
    row0 = mmp.ListOfDouble(values=[1.0, 2.0])
    row1 = mmp.ListOfDouble(values=[2.0, 1.0])
    assert result_dict['test6'] == (
            mmp.SETTING_VALUE_TYPE_LIST_LIST_FLOAT, '', 0, 0.0, False,
            no_list, mmp.ListOfListOfDouble(values=[row0, row1]))


def test_register_instance(mmp_servicer, instance_registry):
    port = mmp.Port(name='test_in', operator=mmp.OPERATOR_F_INIT)
    request = mmp.RegistrationRequest(
            instance_name='test_instance',
            network_locations=['tcp://localhost:10000'],
            ports=[port])

    result = mmp_servicer.RegisterInstance(request, None)

    assert result.status == mmp.RESULT_STATUS_SUCCESS
    assert (instance_registry._InstanceRegistry__locations['test_instance'] ==
            ['tcp://localhost:10000'])

    registered_ports = instance_registry._InstanceRegistry__ports
    assert registered_ports['test_instance'][0].name == 'test_in'
    assert registered_ports['test_instance'][0].operator == Operator.F_INIT


def test_double_register_instance(mmp_servicer, instance_registry):
    port = mmp.Port(name='test_in', operator=mmp.OPERATOR_F_INIT)
    request = mmp.RegistrationRequest(
            instance_name='test_instance',
            network_locations=['tcp://localhost:10000'],
            ports=[port])

    result = mmp_servicer.RegisterInstance(request, None)
    assert result.status == mmp.RESULT_STATUS_SUCCESS

    result = mmp_servicer.RegisterInstance(request, None)
    assert result.status == mmp.RESULT_STATUS_ERROR
    assert 'test_instance' in result.error_message


def test_request_peers_pending(mmp_servicer):
    request = mmp.PeerRequest(instance_name='micro[0][0]')
    result = mmp_servicer.RequestPeers(request, None)
    assert result.status == mmp.RESULT_STATUS_PENDING


def test_request_peers_fanout(registered_mmp_servicer):
    request = mmp.PeerRequest(instance_name='macro')
    result = registered_mmp_servicer.RequestPeers(request, None)
    assert result.status == mmp.RESULT_STATUS_SUCCESS

    assert result.conduits[0] == mmp.Conduit(
            sender='macro.out', receiver='micro.in')
    assert result.conduits[1] == mmp.Conduit(
            sender='micro.out', receiver='macro.in')

    assert result.peer_dimensions[0] == mmp.PeerResult.PeerDimensions(
            peer_name='micro', dimensions=[10, 10])

    for i, peer_locations in enumerate(result.peer_locations):
        instance = 'micro[{}][{}]'.format(i // 10, i % 10)
        location = 'direct:{}'.format(instance)
        assert peer_locations == mmp.PeerResult.PeerLocations(
                instance_name=instance, locations=[location])


def test_request_peers_fanin(registered_mmp_servicer):
    request = mmp.PeerRequest(instance_name='micro[4][3]')
    result = registered_mmp_servicer.RequestPeers(request, None)
    assert result.status == mmp.RESULT_STATUS_SUCCESS

    assert result.conduits[0] == mmp.Conduit(
            sender='macro.out', receiver='micro.in')
    assert result.conduits[1] == mmp.Conduit(
            sender='micro.out', receiver='macro.in')

    assert result.peer_dimensions[0] == mmp.PeerResult.PeerDimensions(
            peer_name='macro', dimensions=[])

    assert result.peer_locations[0] == mmp.PeerResult.PeerLocations(
            instance_name='macro', locations=['direct:macro'])


def test_request_peers_bidir(registered_mmp_servicer2):
    request = mmp.PeerRequest(instance_name='meso[2]')
    result = registered_mmp_servicer2.RequestPeers(request, None)
    assert result.status == mmp.RESULT_STATUS_SUCCESS

    assert result.conduits[0] == mmp.Conduit(
            sender='macro.out', receiver='meso.in')
    assert result.conduits[1] == mmp.Conduit(
            sender='meso.out', receiver='micro.in')
    assert result.conduits[2] == mmp.Conduit(
            sender='micro.out', receiver='meso.in')
    assert result.conduits[3] == mmp.Conduit(
            sender='meso.out', receiver='macro.in')

    assert mmp.PeerResult.PeerDimensions(
            peer_name='micro', dimensions=[5, 10]) in result.peer_dimensions
    assert mmp.PeerResult.PeerDimensions(
            peer_name='macro', dimensions=[]) in result.peer_dimensions

    assert mmp.PeerResult.PeerLocations(
            instance_name='macro',
            locations=['direct:macro']) in result.peer_locations

    for i in range(10):
        instance = 'micro[2][{}]'.format(i)
        location = 'direct:{}'.format(instance)
        assert mmp.PeerResult.PeerLocations(
                instance_name=instance,
                locations=[location]) in result.peer_locations


def test_request_peers_own_conduits(registered_mmp_servicer2):
    request = mmp.PeerRequest(instance_name='macro')
    result = registered_mmp_servicer2.RequestPeers(request, None)
    assert result.status == mmp.RESULT_STATUS_SUCCESS

    assert result.conduits[0] == mmp.Conduit(
            sender='macro.out', receiver='meso.in')
    assert result.conduits[1] == mmp.Conduit(
            sender='meso.out', receiver='macro.in')


def test_request_peers_unknown(registered_mmp_servicer2):
    request = mmp.PeerRequest(instance_name='does_not_exist')
    result = registered_mmp_servicer2.RequestPeers(request, None)
    assert result.status == mmp.RESULT_STATUS_ERROR
    assert result.error_message is not None
    assert 'does_not_exist' in result.error_message



import multiprocessing as mp

import ymmsl

from libmuscle.logging import LogLevel, LogMessage, Timestamp
from libmuscle.manager.instance_registry import InstanceRegistry
from libmuscle.manager.logger import Logger
from libmuscle.manager.mmp_server import MMPServer
from libmuscle.manager.manager import elements_for_model
from libmuscle.manager.topology_store import TopologyStore
from libmuscle.mmp_client import MMPClient
from libmuscle.operator import Operator


def do_logging_test(caplog):
    ymmsl_text = (
            'ymmsl_version: v0.1\n'
            'model:\n'
            '  name: test_model\n'
            '  compute_elements:\n'
            '    macro: macro_implementation\n'
            '    micro:\n'
            '      implementation: micro_implementation\n'
            '      multiplicity: [10]\n'
            '  conduits:\n'
            '    macro.out: micro.in\n'
            '    micro.out: macro.in\n'
            'settings:\n'
            '  test1: 13\n'
            '  test2: 13.3\n'
            '  test3: testing\n'
            '  test4: True\n'
            '  test5: [2.3, 5.6]\n'
            '  test6:\n'
            '    - [1.0, 2.0]\n'
            '    - [3.0, 1.0]\n'
            )

    # create server
    logger = Logger()
    ymmsl_doc = ymmsl.load(ymmsl_text)
    expected_elements = elements_for_model(ymmsl_doc.model)
    instance_registry = InstanceRegistry(expected_elements)
    topology_store = TopologyStore(ymmsl_doc)
    server = MMPServer(logger, ymmsl_doc.settings, instance_registry,
                       topology_store)

    # create client
    client = MMPClient('localhost:9000')
    message = LogMessage(
            instance_id='test_logging',
            timestamp=Timestamp(2.0),
            level=LogLevel.CRITICAL,
            text='Integration testing')

    # log and check
    client.submit_log_message(message)
    assert caplog.records[0].name == 'test_logging'
    assert caplog.records[0].time_stamp == '1970-01-01T00:00:02Z'
    assert caplog.records[0].levelname == 'CRITICAL'
    assert caplog.records[0].message == 'Integration testing'

    server.stop()


def test_logging(log_file_in_tmpdir, caplog):
    process = mp.Process(target=do_logging_test, args=(caplog,))
    process.start()
    process.join()
    assert process.exitcode == 0



from unittest.mock import MagicMock, patch

import pytest
from ymmsl import Conduit, Port, Reference

import muscle_manager_protocol.muscle_manager_protocol_pb2 as mmp

from libmuscle.logging import LogLevel, LogMessage, Timestamp
from libmuscle.mmp_client import MMPClient
from libmuscle.operator import Operator


def test_init() -> None:
    with patch('libmuscle.mmp_client.grpc.insecure_channel'), \
         patch('libmuscle.mmp_client.grpc.channel_ready_future'), \
         patch('muscle_manager_protocol.' +
               'muscle_manager_protocol_pb2_grpc.MuscleManagerStub'
               ) as mock_stub:

        stub = mock_stub.return_value
        client = MMPClient('')
        assert client._MMPClient__client == stub    # type: ignore


def test_connection_fail() -> None:
    with patch('libmuscle.mmp_client.CONNECTION_TIMEOUT', 1):
        with pytest.raises(RuntimeError):
            MMPClient('localhost:9000')


def test_submit_log_message(mocked_mmp_client) -> None:
    message = LogMessage(
            'test_mmp_client',
            Timestamp(1.0),
            LogLevel.WARNING,
            'Testing the MMPClient')

    mocked_mmp_client[0].submit_log_message(message)
    assert mocked_mmp_client[1].SubmitLogMessage.called


def test_get_settings(mocked_mmp_client) -> None:
    client, stub = mocked_mmp_client

    row0 = mmp.ListOfDouble(values=[1.2, 3.4])
    row1 = mmp.ListOfDouble(values=[5.6, 7.8])
    array = mmp.ListOfListOfDouble(values=[row0, row1])
    mmp_values = [
            mmp.Setting(
                name='test1',
                value_type=mmp.SETTING_VALUE_TYPE_STRING,
                value_string='test'),
            mmp.Setting(
                name='test2',
                value_type=mmp.SETTING_VALUE_TYPE_INT,
                value_int=12),
            mmp.Setting(
                name='test3',
                value_type=mmp.SETTING_VALUE_TYPE_FLOAT,
                value_float=3.14),
            mmp.Setting(
                name='test4',
                value_type=mmp.SETTING_VALUE_TYPE_BOOL,
                value_bool=True),
            mmp.Setting(
                name='test5',
                value_type=mmp.SETTING_VALUE_TYPE_LIST_FLOAT,
                value_list_float=mmp.ListOfDouble(values=[1.2, 3.4])),
            mmp.Setting(
                name='test6',
                value_type=mmp.SETTING_VALUE_TYPE_LIST_LIST_FLOAT,
                value_list_list_float=array)]
    settings_result = mmp.SettingsResult(setting_values=mmp_values)
    stub.RequestSettings.return_value = settings_result
    settings = client.get_settings()

    assert len(settings) == 6


def test_register_instance(mocked_mmp_client) -> None:
    mocked_mmp_client[0].register_instance(
            Reference('kernel[13]'),
            ['direct:test', 'tcp:test'],
            [Port('out', Operator.O_I), Port('in', Operator.S)])
    assert mocked_mmp_client[1].RegisterInstance.called


def test_request_peers(mocked_mmp_client) -> None:
    conduits = [mmp.Conduit(sender='kernel.out', receiver='other.in')]
    peer_dimensions = [mmp.PeerResult.PeerDimensions(
            peer_name='other', dimensions=[20])]
    peer_locations = [mmp.PeerResult.PeerLocations(
            instance_name='other', locations=['direct:test', 'tcp:test'])]

    mocked_mmp_client[1].RequestPeers.return_value = mmp.PeerResult(
            status=mmp.RESULT_STATUS_SUCCESS,
            conduits=conduits,
            peer_dimensions=peer_dimensions,
            peer_locations=peer_locations)
    result = mocked_mmp_client[0].request_peers(Reference('kernel[13]'))
    assert mocked_mmp_client[1].RequestPeers.called

    assert len(result[0]) == 1
    assert isinstance(result[0][0], Conduit)
    assert result[0][0].sender == 'kernel.out'
    assert result[0][0].receiver == 'other.in'

    assert isinstance(result[1], dict)
    assert result[1]['other'] == [20]

    assert isinstance(result[2], dict)
    assert result[2]['other'] == ['direct:test', 'tcp:test']


def test_request_peers_error(mocked_mmp_client) -> None:
    mocked_mmp_client[1].RequestPeers.return_value = mmp.PeerResult(
            status=mmp.RESULT_STATUS_ERROR,
            error_message='test_error_message')
    with pytest.raises(RuntimeError):
        mocked_mmp_client[0].request_peers(Reference('kernel[13]'))


def test_request_peers_timeout(mocked_mmp_client) -> None:
    mocked_mmp_client[1].RequestPeers.return_value = mmp.PeerResult(
            status=mmp.RESULT_STATUS_PENDING)
    with patch('libmuscle.mmp_client.PEER_TIMEOUT', 1), \
            patch('libmuscle.mmp_client.PEER_INTERVAL_MIN', 0.1), \
            patch('libmuscle.mmp_client.PEER_INTERVAL_MAX', 1.0):
        with pytest.raises(RuntimeError):
            mocked_mmp_client[0].request_peers(Reference('kernel[13]'))



import logging

from libmuscle.logging import LogLevel, Timestamp
from libmuscle.operator import Operator
from libmuscle.manager.logger import Logger


def test_log_level():
    assert LogLevel.CRITICAL.as_python_level() == logging.CRITICAL
    assert LogLevel.CRITICAL.value > LogLevel.DEBUG.value


def test_create_logger():
    logger = Logger()
    logger.close()


def test_log_message(logger, caplog):
    logger.log_message(
            'test_log_message', Timestamp(123.0),
            LogLevel.CRITICAL, 'Testing the logging system')

    assert caplog.records[0].name == 'test_log_message'
    assert caplog.records[0].time_stamp == '1970-01-01T00:02:03Z'
    assert caplog.records[0].levelname == 'CRITICAL'
    assert caplog.records[0].message == 'Testing the logging system'



from typing import cast

from ymmsl import Reference

from libmuscle.mcp.client import Client
import libmuscle.mcp.pipe_multiplexer as mux


class PipeClient(Client):
    """A client for pipe-based communication.
    """
    @staticmethod
    def can_connect_to(location: str) -> bool:
        """Whether this client class can connect to the given location.

        Args:
            location: The location to potentially connect to.

        Returns:
            True iff this class can connect to this location.
        """
        return mux.can_connect_to(location)

    @staticmethod
    def shutdown(instance_id: Reference) -> None:
        """Free any resources shared by all clients for this instance.
        """
        if mux.can_communicate_for(instance_id):
            # close mux ends to allow clean shutdown
            mux.close_mux_ends(instance_id)
            mux_client_conn = mux.get_instance_client_conn(instance_id)
            mux_client_conn.close()

    def __init__(self, instance_id: Reference, location: str) -> None:
        """Creates a PipeClient.

        Args:
            instance_id: Our instance id.
            peer_id: Id of the peer (server) we're connecting to.
        """
        self._instance_id = instance_id

        mux_client_conn = mux.get_instance_client_conn(instance_id)
        _, peer_id = location.split('/')

        # request connection
        # This assumes that the clients are made one by one in the same thread
        # so that they can use the same pipe without getting in each other's
        # way.
        mux_client_conn.send(peer_id)
        self._conn = mux_client_conn.recv()

    def receive(self, receiver: Reference) -> bytes:
        """Receive a message from a port this client connects to.

        Args:
            receiver: The receiving (local) port.

        Returns:
            The received message.
        """
        self._conn.send(receiver)
        return cast(bytes, self._conn.recv())

    def close(self) -> None:
        """Closes this client.

        This closes any connections this client has and/or performs
        other shutdown activities.
        """
        self._conn.close()



from typing import Dict
import uuid

from ymmsl import Reference

from libmuscle.post_office import PostOffice
from libmuscle.mcp.server import Server


# indexed by server UUID
registered_servers = dict()  # type: Dict[str, Server]


class DirectServer(Server):
    """A server that server within this program

    This server is for connecting compute element instances within the
    same process.
    """
    def __init__(self, instance_id: Reference, post_office: PostOffice
                 ) -> None:
        """Create a DirectServer.

        Args:
            instance_id: Id of the instance we're a server for.
            post_office: A PostOffice to obtain messages from, for
                    servicing requests.
        """
        super().__init__(instance_id, post_office)

        self.__id = str(uuid.uuid4())
        registered_servers[self.__id] = self

    def get_location(self) -> str:
        """Returns the location this server listens on.

        Returns:
            A string containing the location.
        """
        return 'direct:{}'.format(self.__id)

    def request(self, receiver: Reference) -> bytes:
        """Gets the next message for this receiver from its outbox.

        This will block until a message is available.

        Args:
            receiver: The receiver requesting the next message.

        Returns:
            The next message.
        """
        return self._post_office.get_message(receiver)

    def close(self) -> None:
        """Closes this server.

        Since we have no network connections or other resources, this
        is a no-op.
        """
        pass



from typing import cast

from ymmsl import Reference

from libmuscle.mcp.client import Client
from libmuscle.mcp.direct_server import DirectServer, registered_servers


class DirectClient(Client):
    """A client that connects to an MCP server directly.

    This client connects to a DirectServer, and together they provide
    communication within in the same process.
    """
    @staticmethod
    def can_connect_to(location: str) -> bool:
        """Whether DirectClient can connect to the given location.

        Args:
            location: The location to potentially connect to.

        Returns:
            True iff a DirectClient can connect to this location.
        """
        if not location.startswith('direct:'):
            return False
        server_id = location[7:]
        return server_id in registered_servers

    def __init__(self, instance_id: Reference, location: str) -> None:
        """Create a DirectClient for a given location.

        The client will connect to this location and be able to request
        messages from any instance and port represented by it.

        Args:
            instance_id: Our instance id.
            location: A location string.
        """
        super().__init__(instance_id, location)
        server_id = location[7:]
        self.__server = cast(DirectServer, registered_servers[server_id])

    def receive(self, receiver: Reference) -> bytes:
        """Receive a message from a port this client connects to.

        Args:
            receiver: The receiving (local) port.

        Returns:
            The received message.
        """
        return self.__server.request(receiver)

    def close(self) -> None:
        """Closes any resources needed.
        """
        pass



from enum import IntEnum
from typing import Any, cast, Optional

import msgpack

from ymmsl import Reference, Settings


class ExtTypeId(IntEnum):
    """MessagePack extension type ids.

    MessagePack lets you define your own types as an extension to the
    built-in ones. These are distinguished by a number from 0 to 127.
    This class is our registry of extension type ids.
    """
    CLOSE_PORT = 0
    SETTINGS = 1


class ClosePort:
    """Sentinel value to send when closing a port.

    Sending an object of this class on a port/conduit conveys to the
    receiver the message that no further messages will be sent on this
    port during the simulation.

    All information is carried by the type, this has no attributes.
    """
    pass


class Message:
    """A MUSCLE Communication Protocol message.

    Messages carry the identity of their sender and receiver, so that
    they can be routed by a MUSCLE Transport Overlay when we get to
    multi-site running in the future.
    """
    def __init__(self, sender: Reference, receiver: Reference,
                 port_length: Optional[int],
                 timestamp: float, next_timestamp: Optional[float],
                 settings_overlay: Settings, data: Any
                 ) -> None:
        """Create an MCPMessage.

        Senders and receivers are refered to by a Reference, which
        contains Instance[InstanceNumber].Port[Slot].

        The port_length field is only used if two vector ports are
        connected together. In that case the number of slots is not
        determined by the number of instances, and must be set by
        the sender and then communicated to the receiver in this
        additional field in all messages sent on the port.

        Args:
            sender: The sending endpoint.
            receiver: The receiving endpoint.
            port_length: Length of the slot, where applicable.
            settings_overlay: The serialised overlay settings.
            data: The serialised contents of the message.
        """
        self.sender = sender
        self.receiver = receiver
        self.port_length = port_length
        self.timestamp = timestamp
        self.next_timestamp = next_timestamp
        self.settings_overlay = settings_overlay
        self.data = data

    @staticmethod
    def from_bytes(message: bytes) -> 'Message':
        """Create an MCP Message from an encoded buffer.

        Args:
            message: MessagePack encoded message data.
        """
        message_dict = msgpack.unpackb(message, raw=False)
        sender = Reference(message_dict["sender"])
        receiver = Reference(message_dict["receiver"])
        port_length = message_dict["port_length"]
        timestamp = message_dict["timestamp"]
        next_timestamp = message_dict["next_timestamp"]

        settings_dict = msgpack.unpackb(message_dict["settings_overlay"].data,
                                        raw=False)
        settings_overlay = Settings(settings_dict)

        data = message_dict["data"]
        if isinstance(data, msgpack.ExtType):
            if data.code == ExtTypeId.CLOSE_PORT:
                data = ClosePort()
            elif data.code == ExtTypeId.SETTINGS:
                plain_dict = msgpack.unpackb(data.data, raw=False)
                data = Settings(plain_dict)

        return Message(sender, receiver, port_length, timestamp,
                       next_timestamp, settings_overlay, data)

    def encoded(self) -> bytes:
        """Encode the message and return as a bytes buffer.
        """
        # pack overlay
        packed_overlay = msgpack.packb(self.settings_overlay.as_ordered_dict(),
                                       use_bin_type=True)
        overlay = msgpack.ExtType(ExtTypeId.SETTINGS, packed_overlay)

        # pack data
        if isinstance(self.data, ClosePort):
            data = msgpack.ExtType(ExtTypeId.CLOSE_PORT, bytes())
        elif isinstance(self.data, Settings):
            packed_data = msgpack.packb(self.data.as_ordered_dict(),
                                        use_bin_type=True)
            data = msgpack.ExtType(ExtTypeId.SETTINGS, packed_data)
        else:
            data = self.data

        # pack message
        message_dict = {
                'sender': str(self.sender),
                'receiver': str(self.receiver),
                'port_length': self.port_length,
                'timestamp': self.timestamp,
                'next_timestamp': self.next_timestamp,
                'settings_overlay': overlay,
                'data': data
                }

        return cast(bytes, msgpack.packb(message_dict, use_bin_type=True))



from typing import Dict

import pytest
from ymmsl import Reference

from libmuscle.outbox import Outbox
from libmuscle.post_office import PostOffice
from libmuscle.mcp.direct_server import DirectServer, registered_servers
from libmuscle.mcp.message import Message


def test_create(direct_server):
    assert direct_server._DirectServer__id in registered_servers
    assert registered_servers[direct_server._DirectServer__id] == direct_server


def test_location(direct_server):
    assert direct_server.get_location().startswith('direct:')


def test_request(receiver, post_office, direct_server):
    message = Message(Reference('test_sender.test_port'), receiver,
                      None, 0.0, 1.0, bytes(), bytes())
    post_office.outboxes[receiver].deposit(message)
    assert direct_server.request(receiver) == message



import multiprocessing as mp
import os
from pathlib import Path
import pytest
import subprocess
import sys

from libmuscle import Instance, Message
from ymmsl import Operator


def run_macro(instance_id: str):
    sys.argv.append('--muscle-instance={}'.format(instance_id))
    macro()


def macro():
    instance = Instance({
            Operator.O_I: ['out'],
            Operator.S: ['in']})

    while instance.reuse_instance():
        # f_init
        assert instance.get_setting('test1') == 13

        for i in range(2):
            # o_i
            instance.send('out', Message(i * 10.0, (i + 1) * 10.0, 'testing'))

            # s/b
            msg = instance.receive('in')
            assert msg.data == 'testing back {}'.format(i)
            assert msg.timestamp == i * 10.0


def test_cpp_macro_micro(mmp_server_process_simple):
    # only run this if MPI is enabled
    if 'MUSCLE_ENABLE_MPI' not in os.environ:
        pytest.skip('MPI is not enabled, try with MUSCLE_ENABLE_MPI=1')

    # create C++ micro model
    # see libmuscle/cpp/src/libmuscle/tests/micro_model_test.cpp
    cpp_build_dir = Path(__file__).parents[1] / 'libmuscle' / 'cpp' / 'build'
    lib_paths = [
            cpp_build_dir / 'grpc' / 'c-ares' / 'c-ares' / 'lib',
            cpp_build_dir / 'grpc' / 'zlib' / 'zlib' / 'lib',
            cpp_build_dir / 'grpc' / 'openssl' / 'openssl' / 'lib',
            cpp_build_dir / 'protobuf' / 'protobuf' / 'lib',
            cpp_build_dir / 'grpc' / 'grpc' / 'lib',
            cpp_build_dir / 'msgpack' / 'msgpack' / 'lib']
    env = {
            'LD_LIBRARY_PATH': ':'.join(map(str, lib_paths)),
            'PATH': '/usr/bin'}     # allow mpirun to find ssh and not complain
    cpp_test_dir = cpp_build_dir / 'libmuscle' / 'tests'
    mpi_test_micro = cpp_test_dir / 'mpi_micro_model_test'
    micro_result = subprocess.Popen(
            ['mpirun', '-np', '2', '--output-filename', 'mpi_micro.log',
             str(mpi_test_micro), '--muscle-instance=micro'], env=env)

    # run macro model
    macro_process = mp.Process(target=run_macro, args=('macro',))
    macro_process.start()

    # check results
    micro_result.wait()
    assert micro_result.returncode == 0
    macro_process.join()
    assert macro_process.exitcode == 0



from ymmsl import Reference

from libmuscle.mcp.pipe_server import PipeServer
import libmuscle.mcp.pipe_multiplexer as mux


def test_server_shutdown(post_office):
    mux.add_instance('test_instance')
    server = PipeServer(Reference('test_instance'), post_office)
    mux.close_mux_ends('test_instance')
    server.close()
    mux.close_all_pipes()



import socket
from typing import Dict

import msgpack
import pytest
from ymmsl import Reference

from libmuscle.outbox import Outbox
from libmuscle.post_office import PostOffice
from libmuscle.mcp.message import Message


def test_create(tcp_server):
    assert tcp_server._instance_id == Reference('test_sender')
    assert tcp_server._server_thread.is_alive()


def test_location(tcp_server):
    assert tcp_server.get_location().startswith('tcp:')


def test_request(receiver, post_office, tcp_server):
    message = b'testing'
    post_office.outboxes[receiver].deposit(message)

    location = tcp_server._server.server_address
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        sock.connect(location)
        length = len(str(receiver)).to_bytes(8, byteorder='little')
        sock.sendall(length)
        sock.sendall(str(receiver).encode('utf-8'))

        lenbuf = bytearray(8)
        sock.recv_into(lenbuf, 8)
        length = int.from_bytes(lenbuf, 'little')

        databuf = bytearray(length)
        received_count = 0
        while received_count < length:
            bytes_left = length - received_count
            received_count += sock.recv_into(
                    memoryview(databuf)[received_count:], bytes_left)

    assert databuf == message



import pytest
from unittest.mock import patch

from ymmsl import Settings

from libmuscle.communicator import Message
from libmuscle.mmp_client import MMPClient


@pytest.fixture
def mock_grpc_channel():
    with patch('libmuscle.mmp_client.grpc.insecure_channel'):
        yield None


@pytest.fixture
def mocked_mmp_client(mock_grpc_channel):
    with patch('libmuscle.mmp_client.grpc.insecure_channel'), \
         patch('libmuscle.mmp_client.grpc.channel_ready_future'), \
         patch('muscle_manager_protocol.' +
               'muscle_manager_protocol_pb2_grpc.MuscleManagerStub'
               ) as mock_stub:

        stub = mock_stub.return_value
        yield MMPClient(''), stub


@pytest.fixture
def message() -> Message:
    return Message(0.0, None, b'test', Settings())


@pytest.fixture
def message2() -> Message:
    return Message(0.0, None, {'test': 17}, Settings())



import multiprocessing as mp
from pathlib import Path
import subprocess

import ymmsl
from ymmsl import Port, Reference

from libmuscle.logging import LogLevel, LogMessage, Timestamp
from libmuscle.manager.instance_registry import InstanceRegistry
from libmuscle.manager.logger import Logger
from libmuscle.manager.mmp_server import MMPServer
from libmuscle.manager.manager import elements_for_model
from libmuscle.manager.topology_store import TopologyStore
from libmuscle.mmp_client import MMPClient
from libmuscle.operator import Operator


def do_mmp_client_test(caplog):
    ymmsl_text = (
            'ymmsl_version: v0.1\n'
            'model:\n'
            '  name: test_model\n'
            '  compute_elements:\n'
            '    macro: macro_implementation\n'
            '    micro:\n'
            '      implementation: micro_implementation\n'
            '      multiplicity: [10]\n'
            '  conduits:\n'
            '    macro.out: micro.in\n'
            '    micro.out: macro.in\n'
            'settings:\n'
            '  test1: 13\n'
            '  test2: 13.3\n'
            '  test3: testing\n'
            '  test4: True\n'
            '  test5: [2.3, 5.6]\n'
            '  test6:\n'
            '    - [1.0, 2.0]\n'
            '    - [3.0, 1.0]\n'
            )

    # create server
    logger = Logger()
    ymmsl_doc = ymmsl.load(ymmsl_text)
    expected_elements = elements_for_model(ymmsl_doc.model)
    instance_registry = InstanceRegistry(expected_elements)
    topology_store = TopologyStore(ymmsl_doc)
    server = MMPServer(logger, ymmsl_doc.settings, instance_registry,
                       topology_store)

    # mock the deregistration
    removed_instance = None

    def mock_remove(name: Reference):
        nonlocal removed_instance
        removed_instance = name

    instance_registry.remove = mock_remove

    # add some peers
    instance_registry.add(
            Reference('macro'), ['tcp:test3', 'tcp:test4'],
            [Port('out', Operator.O_I), Port('in', Operator.S)])

    # create C++ client
    # it runs through the various RPC calls
    # see libmuscle/cpp/src/libmuscle/tests/mmp_client_test.cpp
    cpp_build_dir = Path(__file__).parents[1] / 'libmuscle' / 'cpp' / 'build'
    lib_paths = [
            cpp_build_dir / 'grpc' / 'c-ares' / 'c-ares' / 'lib',
            cpp_build_dir / 'grpc' / 'zlib' / 'zlib' / 'lib',
            cpp_build_dir / 'grpc' / 'openssl' / 'openssl' / 'lib',
            cpp_build_dir / 'protobuf' / 'protobuf' / 'lib',
            cpp_build_dir / 'grpc' / 'grpc' / 'lib',
            cpp_build_dir / 'msgpack' / 'msgpack' / 'lib']
    env = {
            'LD_LIBRARY_PATH': ':'.join(map(str, lib_paths))}
    cpp_test_dir = cpp_build_dir / 'libmuscle' / 'tests'
    cpp_test_client = cpp_test_dir / 'mmp_client_test'
    result = subprocess.run([str(cpp_test_client)], env=env)

    # check that C++-side checks were successful
    print(result.stderr, flush=True)
    assert result.returncode == 0

    # check submit_log_message
    assert caplog.records[0].name == 'test_logging'
    assert caplog.records[0].time_stamp == '1970-01-01T00:00:02Z'
    assert caplog.records[0].levelname == 'CRITICAL'
    assert caplog.records[0].message == 'Integration testing'

    # check register_instance
    assert (instance_registry.get_locations('micro[3]') ==
            ['tcp:test1', 'tcp:test2'])
    ports = instance_registry.get_ports('micro[3]')
    assert ports[0].name == 'out'
    assert ports[0].operator == Operator.O_F
    assert ports[1].name == 'in'
    assert ports[1].operator == Operator.F_INIT

    # check deregister_instance
    assert removed_instance == 'micro[3]'

    server.stop()


def test_mmp_client(log_file_in_tmpdir, caplog):
    process = mp.Process(target=do_mmp_client_test, args=(caplog,))
    process.start()
    process.join()
    assert process.exitcode == 0



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

import muscle_manager_protocol_pb2 as muscle__manager__protocol__pb2


class MuscleManagerStub(object):
  """The MUSCLE Manager Protocol
  """

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.SubmitLogMessage = channel.unary_unary(
        '/muscle_manager_protocol.MuscleManager/SubmitLogMessage',
        request_serializer=muscle__manager__protocol__pb2.LogMessage.SerializeToString,
        response_deserializer=muscle__manager__protocol__pb2.LogResult.FromString,
        )
    self.SubmitProfileEvents = channel.unary_unary(
        '/muscle_manager_protocol.MuscleManager/SubmitProfileEvents',
        request_serializer=muscle__manager__protocol__pb2.Profile.SerializeToString,
        response_deserializer=muscle__manager__protocol__pb2.ProfileResult.FromString,
        )
    self.RequestSettings = channel.unary_unary(
        '/muscle_manager_protocol.MuscleManager/RequestSettings',
        request_serializer=muscle__manager__protocol__pb2.SettingsRequest.SerializeToString,
        response_deserializer=muscle__manager__protocol__pb2.SettingsResult.FromString,
        )
    self.RegisterInstance = channel.unary_unary(
        '/muscle_manager_protocol.MuscleManager/RegisterInstance',
        request_serializer=muscle__manager__protocol__pb2.RegistrationRequest.SerializeToString,
        response_deserializer=muscle__manager__protocol__pb2.RegistrationResult.FromString,
        )
    self.RequestPeers = channel.unary_unary(
        '/muscle_manager_protocol.MuscleManager/RequestPeers',
        request_serializer=muscle__manager__protocol__pb2.PeerRequest.SerializeToString,
        response_deserializer=muscle__manager__protocol__pb2.PeerResult.FromString,
        )
    self.DeregisterInstance = channel.unary_unary(
        '/muscle_manager_protocol.MuscleManager/DeregisterInstance',
        request_serializer=muscle__manager__protocol__pb2.DeregistrationRequest.SerializeToString,
        response_deserializer=muscle__manager__protocol__pb2.DeregistrationResult.FromString,
        )


class MuscleManagerServicer(object):
  """The MUSCLE Manager Protocol
  """

  def SubmitLogMessage(self, request, context):
    """Sends a logged message to the Manager
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def SubmitProfileEvents(self, request, context):
    """Sends a batch of profiling events to the Manager
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def RequestSettings(self, request, context):
    """Requests the global base settings
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def RegisterInstance(self, request, context):
    """Registers a newly started kernel instance
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def RequestPeers(self, request, context):
    """Requests information about peers
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def DeregisterInstance(self, request, context):
    """Deregisters an instance on shutdown
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_MuscleManagerServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'SubmitLogMessage': grpc.unary_unary_rpc_method_handler(
          servicer.SubmitLogMessage,
          request_deserializer=muscle__manager__protocol__pb2.LogMessage.FromString,
          response_serializer=muscle__manager__protocol__pb2.LogResult.SerializeToString,
      ),
      'SubmitProfileEvents': grpc.unary_unary_rpc_method_handler(
          servicer.SubmitProfileEvents,
          request_deserializer=muscle__manager__protocol__pb2.Profile.FromString,
          response_serializer=muscle__manager__protocol__pb2.ProfileResult.SerializeToString,
      ),
      'RequestSettings': grpc.unary_unary_rpc_method_handler(
          servicer.RequestSettings,
          request_deserializer=muscle__manager__protocol__pb2.SettingsRequest.FromString,
          response_serializer=muscle__manager__protocol__pb2.SettingsResult.SerializeToString,
      ),
      'RegisterInstance': grpc.unary_unary_rpc_method_handler(
          servicer.RegisterInstance,
          request_deserializer=muscle__manager__protocol__pb2.RegistrationRequest.FromString,
          response_serializer=muscle__manager__protocol__pb2.RegistrationResult.SerializeToString,
      ),
      'RequestPeers': grpc.unary_unary_rpc_method_handler(
          servicer.RequestPeers,
          request_deserializer=muscle__manager__protocol__pb2.PeerRequest.FromString,
          response_serializer=muscle__manager__protocol__pb2.PeerResult.SerializeToString,
      ),
      'DeregisterInstance': grpc.unary_unary_rpc_method_handler(
          servicer.DeregisterInstance,
          request_deserializer=muscle__manager__protocol__pb2.DeregistrationRequest.FromString,
          response_serializer=muscle__manager__protocol__pb2.DeregistrationResult.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'muscle_manager_protocol.MuscleManager', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))



from pathlib import Path
import sys
from typing import Generator, List, Optional, cast

from ymmsl import Conduit, Reference

import muscle_manager_protocol.muscle_manager_protocol_pb2 as mmp


def conduit_to_grpc(conduit: Conduit) -> mmp.Conduit:
    """Converts a ymmsl.Conduit to the corresponding mmp.Conduit.

    Args:
        conduit: A conduit.

    Returns:
        The same conduit, but grpc type.
    """
    return mmp.Conduit(sender=str(conduit.sender),
                       receiver=str(conduit.receiver))


def instance_to_kernel(instance: Reference) -> Reference:
    """Extracts the name of the kernel from an instance name.

    Args:
        instance: The name of a kernel instance.

    Returns:
        The name of its kernel.
    """
    i = len(instance)
    while isinstance(instance[i-1], int):
        i -= 1
    return instance[:i]


def instance_indices(instance: Reference) -> List[int]:
    """Extracts the indices from an instance name.

    Args:
        instance: The name of a kernel instance.

    Returns:
        The name of its kernel.
    """
    i = len(instance)
    while isinstance(instance[i-1], int):
        i -= 1

    # Note that the slice operator on References returns a Reference,
    # which we don't want here.
    return [cast(int, instance[j]) for j in range(i, len(instance))]


def generate_indices(dims: List[int]) -> Generator[List[int], None, None]:
    """Generates all indices in a block of the given dimensions.

    Args:
        dims: The dimensions of the block.

    Yields:
        Lists of indices, one for each point in the block.
    """
    index = [0] * len(dims)
    done = False
    while not done:
        yield index
        done = increment_index(index, dims)


def increment_index(index: List[int], dims: List[int]) -> bool:
    """Increments an index.

    Args:
        index: The index to be incremented.
        dims: The dimensions of the block this index is in.

    Returns:
        True iff the index overflowed and is now all zeros again.
    """
    cur = len(index) - 1
    index[cur] += 1
    while index[cur] == dims[cur]:
        index[cur] = 0
        if cur == 0:
            return True
        cur -= 1
        index[cur] += 1
    return False


def extract_log_file_location(filename: str) -> Optional[Path]:
    """Gets the log file location from the command line.

    Extracts the --muscle-log-file=<path> argument to tell the
    MUSCLE library where to write the local log file. This
    function will extract this argument from the command line
    arguments if it is present. If the given path is to a
    directory, <filename> will be written inside of that directory,
    if the path is not an existing directory, then it will be used
    as the name of the log file to write to. If no command line
    argument is given, <filename> will be written in the current
    directory.

    Args:
        filename: Default file name to use.

    Returns:
        Path to the log file to write.
    """
    # Neither getopt, optparse, or argparse will let me pick out
    # just one option from the command line and ignore the rest.
    # So we do it by hand.
    prefix = '--muscle-log-file='
    given_path_str = ''
    for arg in sys.argv[1:]:
        if arg.startswith(prefix):
            given_path_str = arg[len(prefix):]

    if given_path_str == '':
        return Path('.') / filename

    given_path = Path(given_path_str)

    if given_path.is_dir():
        return given_path / filename
    return given_path



import datetime
from enum import Enum
import time
from typing import Dict, List, Optional

import muscle_manager_protocol.muscle_manager_protocol_pb2 as mmp
from ymmsl import Port, Reference

from libmuscle.operator import operator_to_grpc
from libmuscle.port import optional_port_to_grpc
from libmuscle.timestamp import Timestamp


class ProfileEventType(Enum):
    """Profiling event types for MUSCLE 3.

    These match the types in the MUSCLE Manager Protocol, and should \
    be kept identical to those.
    """
    REGISTER = 0
    CONNECT = 4
    DEREGISTER = 1
    SEND = 2
    RECEIVE = 3

    @staticmethod
    def from_grpc(
            event_type: mmp.ProfileEventType
            ) -> 'ProfileEventType':
        """Creates an event type from a gRPC-generated message.

        Args:
            event_type: A profile event type, received from gRPC.

        Returns:
            The same event type, as a ProfileEventType.
        """
        event_type_map = {
                mmp.PROFILE_EVENT_TYPE_REGISTER: ProfileEventType.REGISTER,
                mmp.PROFILE_EVENT_TYPE_CONNECT: ProfileEventType.CONNECT,
                mmp.PROFILE_EVENT_TYPE_DEREGISTER: ProfileEventType.DEREGISTER,
                mmp.PROFILE_EVENT_TYPE_SEND: ProfileEventType.SEND,
                mmp.PROFILE_EVENT_TYPE_RECEIVE: ProfileEventType.RECEIVE
                }  # type: Dict[mmp.ProfileEventType, ProfileEventType]
        return event_type_map[event_type]

    def to_grpc(self) -> mmp.ProfileEventType:
        """Converts the event type to the gRPC generated type.

        Returns:
            The current event type, as the gRPC type.
        """
        event_type_map = {
                ProfileEventType.REGISTER: mmp.PROFILE_EVENT_TYPE_REGISTER,
                ProfileEventType.CONNECT: mmp.PROFILE_EVENT_TYPE_CONNECT,
                ProfileEventType.DEREGISTER: mmp.PROFILE_EVENT_TYPE_DEREGISTER,
                ProfileEventType.SEND: mmp.PROFILE_EVENT_TYPE_SEND,
                ProfileEventType.RECEIVE: mmp.PROFILE_EVENT_TYPE_RECEIVE,
                }  # type: Dict[ProfileEventType, mmp.ProfileEventType]
        return event_type_map[self]


class ProfileEvent:
    """A profile event as used by MUSCLE 3.

    This represents a single measurement of the timing of some event
    that occurred while executing the simulation.

    Args:
        instance_id: The identifier of the instance that generated \
                this message.
        start_time: When the event started (real-world, not \
                simulation time).
        stop_time: When the event ended (real-world, not simulation \
                time).
        event_type: Type of event that was measured.
        port: Port used for sending or receiving, if applicable.
        port_length: Length of that port, if a vector.
        slot: Slot that was sent or received on, if applicable.
        message_size: Size of the message involved, if applicable.

    Attributes:
        instance_id: The identifier of the instance that generated \
                this message.
        start_time: When the event started (real-world, not \
                simulation time).
        stop_time: When the event ended (real-world, not simulation \
                time).
        event_type: Type of event that was measured.
        port: Port used for sending or receiving, if applicable.
        port_length: Length of that port, if a vector.
        slot: Slot that was sent or received on, if applicable.
        message_size: Size of the message involved, if applicable.
    """
    def __init__(
            self,
            instance_id: Reference,
            start_time: Timestamp,
            stop_time: Timestamp,
            event_type: ProfileEventType,
            port: Optional[Port]=None,
            port_length: Optional[int]=None,
            slot: Optional[int]=None,
            message_size: Optional[int]=None
            ) -> None:

        self.instance_id = instance_id
        self.start_time = start_time
        self.stop_time = stop_time
        self.event_type = event_type
        self.port = port
        self.port_length = port_length
        self.slot = slot
        self.message_size = message_size

    def stop(self) -> None:
        """Sets stop_time to the current time.
        """
        self.stop_time = Timestamp(time.time())

    def to_grpc(self) -> mmp.ProfileEvent:
        """Converts the profile event to the gRPC-generated type.

        Returns:
            This profile event, as the gRPC type.
        """
        return mmp.ProfileEvent(
                instance_id=str(self.instance_id),
                start_time=self.start_time.to_grpc(),
                stop_time=self.stop_time.to_grpc(),
                event_type=self.event_type.to_grpc(),
                port=optional_port_to_grpc(self.port),
                port_length=self.port_length,
                slot=self.slot,
                message_size=self.message_size)



import pathlib
from setuptools import find_packages, setup


_here = pathlib.Path(__file__).resolve().parent
_version_file = _here / 'VERSION'
with _version_file.open('r') as f:
    _version = f.read().strip()

# this is a hack, but it works for now and keeps the version number in a single file
_libmuscle_version_file = _here / 'libmuscle' / 'python' / 'libmuscle' / 'version.py'
with _libmuscle_version_file.open('w') as f:
    f.write('__version__ = \'{}\'\n'.format(_version))

setup(
    name='muscle3',
    version=_version,
    description='Version 3 of the MUltiScale Coupling Library and Environment',
    author='Lourens Veen',
    author_email='l.veen@esciencecenter.nl',
    url='https://github.com/multiscale/muscle3',
    license='Apache License 2.0',
    keywords=['multiscale', 'coupling', 'MUSCLE'],
    classifiers=[
        'Development Status :: 1 - Planning',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6'],

    # packages=['muscle_manager', 'muscle_manager_protocol', 'libmuscle', 'libmuscle.mcp'],
    packages=find_packages() + find_packages('libmuscle/python'),
    package_dir={
        'muscle_manager': 'muscle_manager',
        'libmuscle': 'libmuscle/python/libmuscle'
    },
    entry_points={
        'console_scripts': ['muscle_manager=muscle_manager.muscle_manager:manage_simulation']
    },
    python_requires='>=3.5, <4',
    install_requires=[
        'click',
        'grpcio==1.17.1',
        'msgpack',
        'netifaces',
        'protobuf',
        'ruamel.yaml<=0.15.64',
        'ymmsl==0.10.0'          # Also in .travisrc, update there as well
    ],
    setup_requires=[
        'pytest-runner',
        # dependencies for `python setup.py build_sphinx`
        'breathe',
        'sphinx',
        'recommonmark',
        'sphinx-rtd-theme'
    ],
    tests_require=[
        'coverage<5',
        'mypy',
        'pytest>=3.5',
        'pytest-cov',
        'pytest-pep8',
        'pytest-flake8',
        'pytest-mypy'
    ],
    extras_require={
        'dev': [
            'grpcio-tools==1.17.1',
            'mypy-protobuf',
            'sphinx',
            'sphinx_rtd_theme',
            'yapf',
            'isort'
        ]
    },
)

# clean up the version file again
_libmuscle_version_file.unlink()



from libmuscle.mcp.direct_client import DirectClient
from ymmsl import Reference

from unittest.mock import patch


@patch('libmuscle.mcp.direct_client.registered_servers')
def test_can_connect_to(mock_servers) -> None:
    mock_servers.__contains__.return_value = True
    assert DirectClient.can_connect_to('direct:test_server')
    mock_servers.__contains__.assert_called_with('test_server')

    mock_servers.__contains__.return_value = False
    assert not DirectClient.can_connect_to('direct_test_server')


@patch('libmuscle.mcp.direct_client.registered_servers')
def test_create(mock_servers) -> None:
    mock_servers.__getitem__.return_value = 'mock_server'
    client = DirectClient(Reference('test_instance'), 'direct:test_server')
    assert client._DirectClient__server == 'mock_server'


@patch('libmuscle.mcp.direct_client.registered_servers')
def test_receive(mock_servers) -> None:
    client = DirectClient(Reference('test_instance'), 'direct:test_server')
    receiver = Reference('receiver')
    client.receive(receiver)
    client._DirectClient__server.request.assert_called_with(receiver)



from unittest.mock import patch
from typing import Dict

import pytest
from ymmsl import Reference

from libmuscle.mcp.direct_client import DirectClient
from libmuscle.mcp.message import Message


def test_send_receive(receiver, direct_server, post_office):
    assert DirectClient.can_connect_to(direct_server.get_location())
    client = DirectClient(Reference('test_receiver'),
                          direct_server.get_location())

    message = Message(Reference('test_sender.test_port'), receiver,
                      None, 0.0, 1.0, bytes(), 'message'.encode('utf-8'))

    post_office.outboxes[receiver].deposit(message)
    message2 = client.receive(receiver)

    assert message == message2



import pytest

from ymmsl import Operator

from libmuscle.manager.instance_registry import Port, InstanceRegistry


@pytest.fixture
def port():
    return Port('test_port', Operator.F_INIT)


@pytest.fixture
def registry():
    return InstanceRegistry(['instance1'])


def test_port(port):
    assert port.name == 'test_port'
    assert port.operator == Operator.F_INIT


def test_registry_add(registry, port):
    registry.add('instance1', 'tcp://localhost:6253', [port])
    assert (registry._InstanceRegistry__locations['instance1'] ==
            'tcp://localhost:6253')
    assert registry._InstanceRegistry__ports['instance1'] == [port]


def test_registry_get(registry, port):
    registry._InstanceRegistry__locations['instance1'] = [
            'tcp://localhost:6253']
    registry._InstanceRegistry__ports['instance1'] = [port]
    assert registry.get_locations('instance1') == ['tcp://localhost:6253']
    assert registry.get_ports('instance1') == [port]

    with pytest.raises(KeyError):
        registry.get_locations('non-existant-instance')

    with pytest.raises(KeyError):
        registry.get_ports('non-existant-instance')


def test_registry_remove(registry, port):
    registry._InstanceRegistry__locations['instance1'] = [
            'tcp://localhost:6253']
    registry._InstanceRegistry__ports['instance1'] = [port]
    registry.remove('instance1')
    assert 'instance1' not in registry._InstanceRegistry__locations
    assert 'instance1' not in registry._InstanceRegistry__ports

    with pytest.raises(KeyError):
        registry.remove('non-existant-instance')



from typing import List

from ymmsl import Configuration, Model

from libmuscle.manager.instance_registry import InstanceRegistry
from libmuscle.manager.logger import Logger
from libmuscle.manager.mmp_server import MMPServer
from libmuscle.manager.topology_store import TopologyStore


def elements_for_model(model: Model) -> List[str]:
    """Creates a list of elements to expect to register.

    Args:
        model: The model to create a list for.
    """
    def increment(index: List[int], dims: List[int]) -> None:
        # assumes index and dims are the same length > 0
        # modifies index argument
        i = len(index) - 1
        index[i] += 1
        while index[i] == dims[i]:
            index[i] = 0
            i -= 1
            if i == -1:
                break
            index[i] += 1

    def index_to_str(index: List[int]) -> str:
        result = ''
        for i in index:
            result += '[{}]'.format(i)
        return result

    def generate_indices(multiplicity: List[int]) -> List[str]:
        # n-dimensional counter
        indices = list()    # type: List[str]

        index = [0] * len(multiplicity)
        indices.append(index_to_str(index))
        increment(index, multiplicity)
        while sum(index) > 0:
            indices.append(index_to_str(index))
            increment(index, multiplicity)
        return indices

    result = list()     # type: List[str]
    for element in model.compute_elements:
        if len(element.multiplicity) == 0:
            result.append(str(element.name))
        else:
            for index in generate_indices(element.multiplicity):
                result.append(str(element.name) + index)
    return result


def start_server(configuration: Configuration) -> MMPServer:
    """Creates an MMP server and starts it.

    Args;
        configuration: The configuration to run.
    """
    if configuration.settings is None:
        raise ValueError('The yMMSL description needs to specify the'
                         ' settings for the simulation.')
    if not configuration.model or not isinstance(configuration.model, Model):
        raise ValueError('The yMMSL description needs to specify the'
                         ' model to run.')

    topology_store = TopologyStore(configuration)
    expected_elements = elements_for_model(configuration.model)

    logger = Logger()
    instance_registry = InstanceRegistry(expected_elements)
    return MMPServer(logger, configuration.settings, instance_registry,
                     topology_store)



import datetime
from enum import Enum
import logging
from typing import Dict, NewType
from ymmsl import Operator

import muscle_manager_protocol.muscle_manager_protocol_pb2 as mmp

from libmuscle.operator import operator_to_grpc
from libmuscle.timestamp import Timestamp


class LogLevel(Enum):
    """Log levels for MUSCLE 3.

    These match the levels in the MUSCLE Manager Protocol, and should \
    be kept identical to those. They also match the Python logging log \
    levels, although not numerically.
    """
    CRITICAL = 5
    ERROR = 4
    WARNING = 3
    INFO = 1
    DEBUG = 0

    @staticmethod
    def from_python_level(level: int) -> 'LogLevel':
        """Creates a LogLevel from a Python log level.

        Args:
            level: A standard Python log level, such as logging.WARNING.

        Returns:
            The same level, but as a LogLevel.
        """
        if level > 40:
            return LogLevel.CRITICAL
        elif level > 30:
            return LogLevel.ERROR
        elif level > 20:
            return LogLevel.WARNING
        elif level > 10:
            return LogLevel.INFO
        return LogLevel.DEBUG

    def as_python_level(self) -> int:
        """Convert the LogLevel to the corresponding Python level."""
        to_python_level = {
                LogLevel.CRITICAL: logging.CRITICAL,
                LogLevel.ERROR: logging.ERROR,
                LogLevel.WARNING: logging.WARNING,
                LogLevel.INFO: logging.INFO,
                LogLevel.DEBUG: logging.DEBUG}
        return to_python_level[self]

    @staticmethod
    def from_grpc(
            level: mmp.LogLevel
            ) -> 'LogLevel':
        """Creates a log level from a gRPC-generated LogLevel message.

        Args:
            level: A log level, received from gRPC.

        Returns:
            The same log level, as a LogLevel.
        """
        log_level_map = {
                mmp.LOG_LEVEL_DEBUG: LogLevel.DEBUG,
                mmp.LOG_LEVEL_INFO: LogLevel.INFO,
                mmp.LOG_LEVEL_WARNING: LogLevel.WARNING,
                mmp.LOG_LEVEL_ERROR: LogLevel.ERROR,
                mmp.LOG_LEVEL_CRITICAL: LogLevel.CRITICAL
                }  # type: Dict[mmp.LogLevel, LogLevel]
        return log_level_map[level]

    def to_grpc(self) -> mmp.LogLevel:
        """Converts the log level to the gRPC generated type.

        Args:
            level: A log level.

        Returns:
            The same log level, as the gRPC type.
        """
        log_level_map = {
                LogLevel.DEBUG: mmp.LOG_LEVEL_DEBUG,
                LogLevel.INFO: mmp.LOG_LEVEL_INFO,
                LogLevel.WARNING: mmp.LOG_LEVEL_WARNING,
                LogLevel.ERROR: mmp.LOG_LEVEL_ERROR,
                LogLevel.CRITICAL: mmp.LOG_LEVEL_CRITICAL
                }  # type: Dict[LogLevel, mmp.LogLevel]
        return log_level_map[self]


class LogMessage:
    """A log message as used by MUSCLE 3.

    Args:
        instance_id: The identifier of the instance that generated \
                this message.
        timestamp: When the message was generated (real-world, not \
                simulation).
        level: Log level of the message.
        text: Content of the message.

    Attributes:
        instance_id: The identifier of the instance that generated \
                this message.
        timestamp: When the message was generated (real-world, not \
                simulation).
        level: Log level of the message.
        text: Content of the message.
    """
    def __init__(
            self,
            instance_id: str,
            timestamp: Timestamp,
            level: LogLevel,
            text: str
            ) -> None:

        self.instance_id = instance_id
        self.timestamp = timestamp
        self.level = level
        self.text = text

    def to_grpc(self) -> mmp.LogMessage:
        """Converts the log message to the gRPC-generated type.

        Args:
            message: A log message.

        Returns:
            The same log message, as the gRPC type.
        """
        return mmp.LogMessage(
                instance_id=self.instance_id,
                timestamp=self.timestamp.to_grpc(),
                level=self.level.to_grpc(),
                text=self.text)



import multiprocessing as mp
from multiprocessing.connection import Connection
# The below line seems to help avoid crashes, something to do with
# a background thread in the library and forking threaded processes.
from multiprocessing import resource_sharer    # type: ignore
from typing import Dict, List, Tuple
import uuid

from ymmsl import Reference


# This doesn't really multiplex anything, it facilitates the creation
# of pipes between compute element instances running in different
# processes. It's kind of similar to the MUSCLE manager in that, but
# naming it manager would be even more confusing, so multiplexer it is.


class _InstancePipe:
    """Pipes for communicating between an instance and the mux.

    Objects of this class contain the endpoints for a pipe
    that is used to communicate between an instance processe and the
    multiplexer. The multiplexer (in this module) facilitates the
    creation of peer-to-peer pipe connections between the processes.

    For each instance there is a server pipe, on which the instance
    listens for incoming connection requests, and a client pipe,
    through which an instance asks the multiplexer to create a
    connection to another instance.

    Attributes:
        mux_conn: Mux side of the pipe.
        instance_conn: Instance side of the pipe.
    """
    def __init__(self) -> None:
        """Create an InstancePipes containing two new pipes.
        """
        self.mux_conn, self.instance_conn = mp.Pipe()


_instance_client_pipes = dict()     # type: Dict[Reference, _InstancePipe]
_instance_server_pipes = dict()     # type: Dict[Reference, _InstancePipe]


def add_instance(instance_id: Reference) -> None:
    """Adds pipes for an instance.

    Args:
        instance_id: Name of the new instance.
    """
    _instance_client_pipes[instance_id] = _InstancePipe()
    _instance_server_pipes[instance_id] = _InstancePipe()


def can_communicate_for(instance_id: Reference) -> bool:
    """Returns whether we can serve on pipes for this instance.

    If the instance was not started via Muscle3, then it will not be
    registered, and we cannot start a PipeServer for it. That's fine,
    but we need to know.

    Args:
        instance_id: Name of the requested instance.
    """
    return instance_id in _instance_server_pipes


def close_instance_ends(instance_id: Reference) -> None:
    """Closes the instance sides of the pipes for the given instance.

    Args:
        instance_id: The instance to close for.
    """
    _instance_client_pipes[instance_id].instance_conn.close()
    _instance_server_pipes[instance_id].instance_conn.close()


def close_mux_ends(instance_id: Reference) -> None:
    """Closes the mux sides of the pipes for the given instance.

    Args:
        instance_id: The instance to close for.
    """
    _instance_client_pipes[instance_id].mux_conn.close()
    _instance_server_pipes[instance_id].mux_conn.close()


def close_all_pipes() -> None:
    """Closes all the instance pipes.

    For example, because each instance process and the mux process have
    their own copies, so the main process doesn't need them and its
    copies should be closed.
    """
    def close_all(pipes: Dict[Reference, _InstancePipe]) -> None:
        instances = list()  # type: List[Reference]
        for instance_id, instance_pipe in pipes.items():
            instance_pipe.instance_conn.close()
            instance_pipe.mux_conn.close()
            instances.append(instance_id)

        for instance_id in instances:
            del(pipes[instance_id])

    close_all(_instance_client_pipes)
    close_all(_instance_server_pipes)


def get_instance_server_conn(instance_id: Reference) -> Connection:
    """Returns the instance side of the server pipe for this instance.

    Args:
        instance_id: The instance for which to get the pipe.

    Returns:
        The instance side of the server pipe.
    """
    return _instance_server_pipes[instance_id].instance_conn


def get_instance_client_conn(instance_id: Reference) -> Connection:
    """Returns the instance side of the client pipe for this instance.

    Args:
        instance_id: The instance for which to get the pipe.

    Returns:
        The instance side of the client pipe.
    """
    return _instance_client_pipes[instance_id].instance_conn


_process_uuid = uuid.uuid4()


def get_address_for(instance_id: Reference) -> str:
    """Returns the MUSCLE-address for the given instance.

    Args:
        instance_id: Id of the instance to get the address for.

    Returns:
        An address string that can be passed to the MUSCLE Manager.
    """
    return 'pipe:{}/{}'.format(_process_uuid, instance_id)


def can_connect_to(peer_address: str) -> bool:
    """Checks whether this multiplexer can make connections for the address.

    This does not check whether the peer is up and running and
    listening, just that it's in the right process tree.

    Args:
        peer_address: An address of the form <uuid>/<instance_id>.
    """
    if not peer_address.startswith('pipe:'):
        return False
    peer_uuid = peer_address[5:].split('/')[0]
    return peer_uuid == str(_process_uuid)


def run() -> None:
    """Runs the pipe-based communication multiplexer.

    This listens for connection requests from instances, and creates
    pipes between the requested instances so that they can exchange
    messages.

    Shuts down automatically when all client instances have shut down.
    """
    while len(_instance_client_pipes) > 0:
        client_pipes = {instance_id: pipes.mux_conn
                        for instance_id, pipes
                        in _instance_client_pipes.items()}
        ready_pipes = mp.connection.wait(client_pipes.values())

        for client_id, pipe in client_pipes.items():
            if pipe in ready_pipes:
                try:
                    requested_server = pipe.recv()
                    conn1, conn2 = mp.Pipe()
                    _instance_server_pipes[requested_server].mux_conn.send(
                            (conn1, client_id))
                    pipe.send(conn2)
                    # close our copy of this pipe, they'll do peer-to-peer
                    conn1.close()
                    conn2.close()
                except EOFError:
                    _instance_client_pipes[client_id].mux_conn.close()
                    del(_instance_client_pipes[client_id])

    server_ids = list(_instance_server_pipes.keys())
    for instance_id in server_ids:
        _instance_server_pipes[instance_id].mux_conn.close()
        del(_instance_server_pipes[instance_id])

    # Python uses a background thread to help share file descriptors over
    # pipes. The below statement stops that thread, so that it doesn't
    # cause crashes when we fork in later tests.
    resource_sharer.stop()



from unittest.mock import patch

import pytest
from ymmsl import Conduit, Operator, Port, Reference

from libmuscle.mmp_client import MMPClient


def test_registration(log_file_in_tmpdir, mmp_server):
    client = MMPClient('localhost:9000')
    instance_name = Reference('test_instance')
    port = Port(Reference('test_in'), Operator.S)

    client.register_instance(instance_name, ['tcp://localhost:10000'],
                             [port])

    servicer = mmp_server._MMPServer__servicer
    registry = servicer._MMPServicer__instance_registry

    assert registry.get_locations(instance_name) == ['tcp://localhost:10000']
    assert registry.get_ports(instance_name)[0].name == 'test_in'
    assert registry.get_ports(instance_name)[0].operator == Operator.S


def test_wiring(log_file_in_tmpdir, mmp_server_process):
    client = MMPClient('localhost:9000')

    client.register_instance(Reference('macro'), ['direct:macro'], [])

    conduits, peer_dims, peer_locations = client.request_peers(
            Reference('micro[0]'))

    assert Conduit('macro.out', 'micro.in') in conduits
    assert Conduit('micro.out', 'macro.in') in conduits

    assert peer_dims[Reference('macro')] == []
    assert peer_locations['macro'] == ['direct:macro']

    with patch('libmuscle.mmp_client.PEER_TIMEOUT', 0.1), \
            patch('libmuscle.mmp_client.PEER_INTERVAL_MIN', 0.01), \
            patch('libmuscle.mmp_client.PEER_INTERVAL_MAX', 0.1):
        with pytest.raises(RuntimeError):
            client.request_peers(Reference('macro'))

    for i in range(5):
        instance = Reference('micro[{}]'.format(i))
        location = 'direct:{}'.format(instance)
        client.register_instance(instance, [location], [])

    with patch('libmuscle.mmp_client.PEER_TIMEOUT', 0.1), \
            patch('libmuscle.mmp_client.PEER_INTERVAL_MIN', 0.01), \
            patch('libmuscle.mmp_client.PEER_INTERVAL_MAX', 0.1):
        with pytest.raises(RuntimeError):
            client.request_peers(Reference('macro'))

    for i in range(5, 10):
        instance = Reference('micro[{}]'.format(i))
        location = 'direct:{}'.format(instance)
        client.register_instance(instance, [location], [])

    conduits, peer_dims, peer_locations = client.request_peers(
            Reference('macro'))

    assert Conduit('macro.out', 'micro.in') in conduits
    assert Conduit('micro.out', 'macro.in') in conduits

    assert peer_dims[Reference('micro')] == [10]
    assert peer_locations['micro[7]'] == ['direct:micro[7]']



import multiprocessing as mp
from multiprocessing.connection import Connection
import threading
from ymmsl import Reference

from libmuscle.post_office import PostOffice
import libmuscle.mcp.pipe_multiplexer as mux
from libmuscle.mcp.server import Server, ServerNotSupported


class PipeServer(Server):
    """A server that accepts MCP connections over pipes.

    This server uses the pipe_multiplexer module to service message
    requests arriving via a pipe to another process containing a
    compute element instance.
    """
    def __init__(self, instance_id: Reference, post_office: PostOffice
                 ) -> None:
        """Create a PipeServer.

        Args:
            instance_id: Id of the instance for which we're serving.
        """
        super().__init__(instance_id, post_office)
        if mux.can_communicate_for(instance_id):
            # close mux ends to allow clean shutdown
            mux.close_mux_ends(instance_id)

            self._mux_server_conn = mux.get_instance_server_conn(instance_id)
            self._shutdown_conn, self._handler_shutdown_conn = mp.Pipe()
            self._server_thread = threading.Thread(
                    target=self.__conn_request_handler,
                    name='PipeServer-{}'.format(instance_id))

            self._server_thread.start()
        else:
            raise ServerNotSupported(('Instance {} was not started via'
                                      ' Muscle3, cannot communicate via pipes'
                                      ' because the required set-up was not'
                                      ' done.').format(self._instance_id))

    def get_location(self) -> str:
        """Returns the location this server listens on.

        Returns:
            A string containing the location.
        """
        return mux.get_address_for(self._instance_id)

    def close(self) -> None:
        """Closes this server.

        Stops the server listening, waits for existing clients to
        disconnect, then frees any other resources.
        """
        self._shutdown_conn.close()
        self._server_thread.join()

    def __conn_request_handler(self) -> None:
        """Handles incoming connection requests from the mux.
        """
        conn_threads = list()
        while True:
            ready_conns = mp.connection.wait(
                    [self._mux_server_conn, self._handler_shutdown_conn])

            if self._mux_server_conn in ready_conns:
                try:
                    connection, client_id = self._mux_server_conn.recv()
                    conn_thread = threading.Thread(
                            target=self.__mcp_pipe_handler,
                            args=(client_id, connection),
                            name='PipeHandler-{}-{}'.format(
                                self._instance_id, client_id))
                    conn_thread.start()
                    conn_threads.append(conn_thread)
                except EOFError:
                    self._handler_shutdown_conn.close()
                    break

            if self._handler_shutdown_conn in ready_conns:
                self._handler_shutdown_conn.close()
                break

        self._mux_server_conn.close()

        # Wait for clients to disconnect
        for conn_thread in conn_threads:
            conn_thread.join()

    def __mcp_pipe_handler(self, client_id: Reference, connection: Connection
                           ) -> None:
        """Handles a single connection from a single client.

        Args:
            client_id: Id of the client we're talking to.
            connection: The connection with the client.
        """
        try:
            while True:
                receiver_id = connection.recv()
                msg = self._post_office.get_message(receiver_id)
                connection.send(msg)
        except EOFError:
            pass



from pathlib import Path
import subprocess
from unittest.mock import MagicMock

import msgpack

from libmuscle.mcp.tcp_client import TcpClient
from libmuscle.mcp.message import Message

from ymmsl import Reference, Settings


def test_cpp_tcp_server(log_file_in_tmpdir):
    # create C++ server
    # it serves a message for us to receive
    # see libmuscle/cpp/src/libmuscle/tests/mmp_server_test.cpp
    cpp_build_dir = Path(__file__).parents[1] / 'libmuscle' / 'cpp' / 'build'
    lib_paths = [
            cpp_build_dir / 'grpc' / 'c-ares' / 'c-ares' / 'lib',
            cpp_build_dir / 'grpc' / 'zlib' / 'zlib' / 'lib',
            cpp_build_dir / 'grpc' / 'openssl' / 'openssl' / 'lib',
            cpp_build_dir / 'protobuf' / 'protobuf' / 'lib',
            cpp_build_dir / 'grpc' / 'grpc' / 'lib',
            cpp_build_dir / 'msgpack' / 'msgpack' / 'lib']
    env = {
            'LD_LIBRARY_PATH': ':'.join(map(str, lib_paths))}
    cpp_test_dir = cpp_build_dir / 'libmuscle' / 'tests'
    cpp_test_server = cpp_test_dir / 'tcp_server_test'
    server = subprocess.Popen(
            [str(cpp_test_server)], env=env, stdin=subprocess.DEVNULL,
            stdout=subprocess.PIPE,
            universal_newlines=True, bufsize=1, close_fds=True)

    # get server location
    location = server.stdout.readline()

    assert TcpClient.can_connect_to(location)

    client = TcpClient(Reference('test_receiver'), location)
    msg = Message.from_bytes(client.receive(Reference('test_receiver.port')))
    client.close()

    # assert stuff
    assert msg.sender == 'test_sender.port'
    assert msg.receiver == 'test_receiver.port'
    assert msg.timestamp == 0.0
    assert msg.next_timestamp == 1.0
    assert msg.settings_overlay == Settings({'par1': 13})
    assert msg.data == {'var1': 1, 'var2': 2.0, 'var3': '3'}

    server.stdout.close()
    server.wait()
    assert server.returncode == 0



import sys
from typing import Generator, List
from unittest.mock import MagicMock, patch

import pytest
from ymmsl import Conduit, Operator, Reference, Settings

from libmuscle.communicator import Message
from libmuscle.instance import Instance
from libmuscle.mcp.message import ClosePort
from libmuscle.settings_manager import SettingsManager


@pytest.fixture
def sys_argv_manager() -> Generator[None, None, None]:
    old_argv = sys.argv
    sys.argv = sys.argv + ['--muscle-manager=localhost:9000']
    yield None
    sys.argv = old_argv


@pytest.fixture
def log_file_in_tmpdir(tmpdir) -> Generator[None, None, None]:
    old_argv = sys.argv
    sys.argv = sys.argv + ['--muscle-log-file={}'.format(tmpdir)]
    yield None
    sys.argv = old_argv


@pytest.fixture
def sys_argv_instance() -> Generator[None, None, None]:
    old_argv = sys.argv
    sys.argv = ['', '--muscle-instance=test_instance[13][42]']
    yield
    sys.argv = old_argv


@pytest.fixture
def instance(sys_argv_instance):
    with patch('libmuscle.instance.MMPClient') as mmp_client, \
         patch('libmuscle.instance.Communicator') as comm_type:
        communicator = MagicMock()
        settings = Settings()
        settings['test1'] = 12
        msg = Message(0.0, 1.0, 'message', settings)
        communicator.receive_message.return_value = msg
        comm_type.return_value = communicator

        mmp_client_object = MagicMock()
        mmp_client_object.request_peers.return_value = (None, None, None)
        mmp_client.return_value = mmp_client_object

        instance = Instance({
            Operator.F_INIT: ['in', 'not_connected'],
            Operator.O_F: ['out']})
        instance._f_init_cache = dict()
        instance._f_init_cache[('in', None)] = msg
        yield instance


@pytest.fixture
def instance2(sys_argv_instance):
    with patch('libmuscle.instance.MMPClient') as mmp_client, \
         patch('libmuscle.instance.Communicator') as comm_type:
        mmp_client_object = MagicMock()
        mmp_client_object.request_peers.return_value = (None, None, None)
        mmp_client.return_value = mmp_client_object
        instance = Instance({
            Operator.F_INIT: ['in[]'],
            Operator.O_F: ['out']})
        yield instance


def test_create_instance(
        sys_argv_instance, log_file_in_tmpdir, sys_argv_manager):
    with patch('libmuscle.instance.MMPClient') as mmp_client, \
         patch('libmuscle.instance.Communicator') as comm_type:
        mmp_client_object = MagicMock()
        mmp_client_object.request_peers.return_value = (None, None, None)
        mmp_client.return_value = mmp_client_object
        ports = {
            Operator.F_INIT: ['in'],
            Operator.O_F: ['out']}
        instance = Instance(ports)
        assert instance._name == Reference('test_instance')
        assert instance._index == [13, 42]
        assert instance._declared_ports == ports
        assert isinstance(instance._settings_manager, SettingsManager)
        assert len(instance._settings_manager.base) == 0
        assert len(instance._settings_manager.overlay) == 0
        mmp_client.assert_called_once_with('localhost:9000')
        assert mmp_client_object._register.called_with()
        assert mmp_client_object._connect.called_with()
        comm_type.assert_called_with(Reference('test_instance'), [13, 42],
                                     ports, instance._profiler)
        assert instance._communicator == comm_type.return_value
        assert isinstance(instance._settings_manager, SettingsManager)
        assert len(instance._settings_manager.base) == 0


def test_extract_manager_location(sys_argv_manager) -> None:
    assert (Instance._Instance__extract_manager_location() ==
            'localhost:9000')


def test_get_setting(instance):
    ref = Reference
    settings = Settings()
    settings[ref('test1')] = 'test'
    settings[ref('test2')] = 12
    settings[ref('test3')] = 27.1
    settings[ref('test4')] = True
    settings[ref('test5')] = [2.3, 5.6]
    settings[ref('test6')] = [[1.0, 2.0], [3.0, 4.0]]
    instance._settings_manager.base = settings

    assert instance.get_setting('test1') == 'test'
    assert instance.get_setting('test2') == 12
    assert instance.get_setting('test3') == 27.1
    assert instance.get_setting('test4') is True
    assert instance.get_setting('test5') == [2.3, 5.6]
    assert instance.get_setting('test6') == [
            [1.0, 2.0], [3.0, 4.0]]

    assert instance.get_setting('test1', 'str') == 'test'
    assert instance.get_setting('test2', 'int') == 12
    assert instance.get_setting('test3', 'float') == 27.1
    assert instance.get_setting('test4', 'bool') is True
    assert (instance.get_setting('test5', '[float]') ==
            [2.3, 5.6])
    assert (instance.get_setting('test6', '[[float]]') ==
            [[1.0, 2.0], [3.0, 4.0]])

    with pytest.raises(KeyError):
        instance.get_setting('testx')

    with pytest.raises(TypeError):
        instance.get_setting('test1', 'int')
    with pytest.raises(TypeError):
        instance.get_setting('test6', '[float]')
    with pytest.raises(TypeError):
        instance.get_setting('test5', '[[float]]')
    with pytest.raises(ValueError):
        instance.get_setting('test2', 'nonexistenttype')


def test_list_ports(instance):
    ports = instance.list_ports()
    assert instance._communicator.list_ports.called_with()
    assert ports == instance._communicator.list_ports.return_value


def test_is_vector_port(instance):
    instance._communicator.get_port.return_value.is_vector = MagicMock(
            return_value=True)
    is_vector = instance.is_vector_port('out_port')
    assert is_vector is True
    assert instance._communicator.get_port.called_with('out_port')


def test_send(instance, message):
    instance.send('out', message, 1)
    assert instance._communicator.send_message.called_with(
            'out', message, 1)


def test_send_invalid_port(instance, message):
    instance._communicator.port_exists.return_value = False
    with pytest.raises(RuntimeError):
        instance.send('does_not_exist', message, 1)


def test_receive(instance):
    instance._communicator.get_port.return_value = MagicMock(
            operator=Operator.F_INIT)
    msg = instance.receive('in')
    assert msg.timestamp == 0.0
    assert msg.next_timestamp == 1.0
    assert instance._communicator.receive_message.called_with(
            'in', None)
    assert msg.data == 'message'

    with pytest.raises(RuntimeError):
        instance.receive('in')


def test_receive_default(instance):
    instance._communicator.port_exists.return_value = True
    port = instance._communicator.get_port.return_value
    port.operator = Operator.F_INIT
    port.is_connected.return_value = False
    instance.receive('not_connected', 1, 'testing')
    assert instance._communicator.receive_message.called_with(
            'not_connected', 1, 'testing')
    with pytest.raises(RuntimeError):
        instance.receive('not_connected', 1)


def test_receive_invalid_port(instance):
    instance._communicator.port_exists.return_value = False
    with pytest.raises(RuntimeError):
        instance.receive('does_not_exist', 1)


def test_receive_with_settings(instance):
    msg = instance.receive_with_settings('in', 1)
    assert (instance._communicator.receive_message
            .called_with('in', 1))
    assert msg.timestamp == 0.0
    assert msg.next_timestamp == 1.0
    assert msg.data == 'message'
    assert msg.settings['test1'] == 12


def test_receive_with_settings_default(instance):
    instance.receive_with_settings('not_connected', 1, 'testing')
    assert instance._communicator.receive_message.called_with(
            'not_connected', 1, 'testing')


def test_receive_parallel_universe(instance) -> None:
    instance._settings_manager.overlay['test2'] = 'test'
    with pytest.raises(RuntimeError):
        instance.receive('in')


def test_reuse_instance_receive_overlay(instance):
    instance._settings_manager.overlay = Settings()
    test_base_settings = Settings()
    test_base_settings['test1'] = 24
    test_base_settings['test2'] = [1.3, 2.0]
    test_overlay = Settings()
    test_overlay['test2'] = 'abc'
    recv = instance._communicator.receive_message
    recv.return_value = Message(0.0, None, test_overlay, test_base_settings)
    instance.reuse_instance()
    assert instance._communicator.receive_message.called_with(
        'muscle_settings_in')
    assert len(instance._settings_manager.overlay) == 2
    assert instance._settings_manager.overlay['test1'] == 24
    assert instance._settings_manager.overlay['test2'] == 'abc'


def test_reuse_instance_closed_port(instance):
    def receive_message(port_name, slot=None, default=None):
        if port_name == 'muscle_settings_in':
            return Message(0.0, None, Settings(), Settings())
        elif port_name == 'in':
            return Message(0.0, None, ClosePort(), Settings())
        assert False    # pragma: no cover

    def get_port(port_name):
        port = MagicMock()
        port.is_vector.return_value = False
        if port_name == 'not_connected':
            port.is_connected.return_value = False
        else:
            port.is_connected.return_value = True
        return port

    instance._communicator.receive_message = receive_message
    instance._communicator.list_ports.return_value = {
            Operator.F_INIT: ['in', 'not_connected'],
            Operator.O_F: ['out']}
    instance._communicator.get_port = get_port
    instance._Instance__close_ports = MagicMock()

    do_reuse = instance.reuse_instance()
    assert do_reuse is False


def test_reuse_instance_vector_port(instance2):
    def receive_message(port_name, slot=None, default=None):
        if port_name == 'muscle_settings_in':
            return Message(0.0, None, Settings(), Settings())
        elif port_name == 'in':
            data = 'test {}'.format(slot)
            return Message(0.0, None, data, Settings())
        assert False    # pragma: no cover

    instance2._communicator.receive_message = receive_message
    instance2._communicator.list_ports.return_value = {
            Operator.F_INIT: ['in'],
            Operator.O_F: ['out']}

    port = MagicMock()
    port.is_vector.return_value = True
    port.is_connected.return_value = True
    port.get_length.return_value = 10
    instance2._communicator.get_port.return_value = port

    do_reuse = instance2.reuse_instance()
    assert do_reuse is True

    msg = instance2.receive('in', 5)
    assert msg.timestamp == 0.0
    assert msg.next_timestamp is None
    assert msg.data == 'test 5'


def test_reuse_instance_no_f_init_ports(instance):
    instance._communicator.receive_message.return_value = Message(
            0.0, None, Settings(), Settings())
    instance._communicator.list_ports.return_value = {}
    instance._communicator.settings_in_connected.return_value = False
    do_reuse = instance.reuse_instance()
    assert do_reuse is True
    do_reuse = instance.reuse_instance()
    assert do_reuse is False


def test_reuse_instance_miswired(instance):
    with pytest.raises(RuntimeError):
        instance.reuse_instance()
